(* ************************************************** *)
(*                                                    *)
(*    RAW ABSTRACT SYNTAX TREES WITH LOCATION INFO    *)
(*                                                    *)
(* ************************************************** *)

open Globals
open Library
open Location
open Machine
open Types
open Errors

(* ******************* *)
(*   RAW IDENTIFIERS   *)
(* ******************* *)

type rawId = string * span

(* Convert a rawId into a normal string *)
let getStr (x,sp) = x

(* Raw integer expression, for projection *)
type rawInt = int * span

(* Convert rawInt into a normal int *)
let getInt (i,sp) = i

(* The string which is used for implementing wildcard "patterns".
 * This works because this string would be a lexical error in a user program. *)
let wildcard = "%_"

(* Compare two rawIds, just considering the string component. *)
let cmpId x y = (getStr x) = (getStr y)

(* ************* *)
(*   RAW TYPES   *)
(* ************* *)

(* Raw alphaML types - need to resolve type names into either
 * datatypes or nametypes *)
type rawType = 
  RawUnitTy
| RawPropTy
| RawRelTy
| RawTypeId of string
| RawTupTy  of rawType list
| RawFunTy  of rawType * rawType
| RawAbsTy  of string * rawType

(* to_string function for raw types *)
let rec soRawType x = 
 let brackets x = match x with
   RawTupTy _ | RawFunTy _ -> "(" ^ (soRawType x) ^ ")"
 | _ -> soRawType x
 in match x with
  RawUnitTy -> "unit"
| RawPropTy -> "prop"
| RawRelTy -> "rel"
| RawTypeId x -> x
| RawTupTy rts -> string_of_list brackets " * " rts
| RawFunTy (rt,rt') -> (brackets rt) ^ " -> " ^ (brackets rt')
| RawAbsTy (x,rt) -> "[" ^ x ^ "]" ^ (brackets rt)

(* Classify type identifiers according to the provided lists of
 * nametypes and datatypes, which are assumed to be distinct.
 * An exception is raised if an identifier cannot be interpreted. *)
let classifyIds ntys dtys x = if (List.mem x ntys) then Nty x
  else (if (List.mem x dtys) then Dty x
        else raise (UnboundTypeId x))

(* Translate a rawType into a Types.ty, using the supplied
 * function to interpret any type identifiers. *)
let rec trnRawType f = function
  RawUnitTy -> UnitTy
| RawPropTy -> PropTy
| RawRelTy -> RelTy
| RawTypeId x -> f x
| RawTupTy rts -> TupTy (List.map (trnRawType f) rts)
| RawFunTy(rt,rt') -> FunTy((trnRawType f rt),(trnRawType f rt'))
| RawAbsTy(x,rt) -> (match f x with
    Nty n -> AbsTy(n,(trnRawType f rt))
  | _ -> raise (AbsError x))

(* ************************* *)
(*   RAW TYPE DECLARATIONS   *)
(* ************************* *)

(* Raw datatype declaration, as returned from parser *)
type rawDtyDec = (string * ((string * rawType) list)) list

(* to_string function for raw datatype declarations *)
let rawdec_toString (rdec:rawDtyDec) = match rdec with
  [] -> ""
| d::rst ->
    let string_of_line initial (d,kts) =
      let shift = (if (initial="datatype ") then 9 else 4) in
      let sep = ("\n" ^ (String.make ((String.length d)+shift) ' ') ^ " | ") in
      (initial ^ d ^ " = " ^
	 (string_of_list (fun (k,t) -> k ^ " of " ^ (soRawType t)) sep kts))
    in
    ((string_of_line "datatype " d) ^
     (string_of_list (string_of_line "\nand ") "" rst))

(* Raw relation declaration, as returned from parser *)
type rawRelDec = (string * rawType) list

(* to_string function for raw relation declarations *)
let rawRelDec_toString (rdec:rawRelDec) =
  let string_of_line (r,rt) = (r ^ " <: " ^ (soRawType rt)) in
  ("relation " ^ (string_of_list string_of_line "\n     and " rdec))

(* ******************************************** *)
(*   RAW EXPRESSIONS AND INDUCTIVE DEFINITIONS  *)
(* *******************************************  *)

(* Raw ground trees. *)
type rawTree =
  RawNameTree of rawId
| RawUnitTree of span
| RawAbsTree of rawId * rawType * rawTree * span
| RawDataTree of rawId * rawTree * span
| RawTupTree of (rawTree list) * span

(* Raw expressions *)
type rawExpr =
  RawVarExpr of rawId
| RawEvarExpr of rawId
| RawLamExpr of rawId * rawType * rawExpr * span
| RawRecExpr of rawId * rawId * rawType * rawType * rawExpr * span
| RawAppExpr of (rawExpr list) * span
| RawProjExpr of rawExpr * rawInt * span
| RawCaseExpr of rawExpr * ((rawId * rawId * rawExpr) list) * span
| RawDataExpr of rawId * rawExpr * span
| RawTupExpr of (rawExpr list) * span
| RawUnitExpr of span
| RawSomeExpr of rawType * span
| RawAbsExpr of rawExpr * rawExpr * span
| RawBranchExpr of (rawExpr list) * span
| RawSeqExpr of rawExpr * rawExpr * span
| RawFailExpr of rawType * span
| RawSuccessExpr of span
| RawEconExpr of rawExpr * rawExpr * span
| RawFconExpr of rawExpr * rawExpr * span
| RawFreshExpr of rawId * span
| RawDistinctExpr of (rawExpr list) * span
| RawDefnExpr of rawDefn * span
| RawLetExpr of rawId * (rawId * rawType) list * rawExpr * rawExpr * span
| RawLetRecExpr of rawId * (rawId * rawType) list * rawType * rawExpr * rawExpr * span
| RawUnbindExpr of rawExpr * rawId * rawId * rawId * rawType * rawExpr * bool * span
| RawExistsExpr of rawId * rawType * rawExpr * span
| RawTreeExpr of rawTree * span
| RawPrintExpr of string * span
(* Raw patterns *)
and rawPat =
  RawVarPat of rawId  
| RawUnitPat of span
| RawAbsPat of rawId * rawPat * span
| RawTupPat of (rawPat list) * span
| RawDataPat of rawId * rawPat * span
(* Raw rules and definitions *)
and rawAtomic = rawId * ((rawPat list) * span) * span
and rawRule =
  rawExpr * rawAtomic * (string option) * ((rawId * rawType) list) * span
and rawDefn = rawRule list

(* Get span information out of ASTs *)
let getSpanId (_,sp) = sp
let getSpanInt (_,sp) = sp
let getSpanExpr = function
  RawVarExpr x | RawEvarExpr x -> getSpanId x
| RawLamExpr (_,_,_,sp) | RawRecExpr (_,_,_,_,_,sp)
| RawAppExpr (_,sp) | RawLetExpr (_,_,_,_,sp)
| RawProjExpr (_,_,sp) | RawCaseExpr (_,_,sp)
| RawDataExpr (_,_,sp) | RawTupExpr (_,sp)
| RawUnitExpr sp | RawSomeExpr (_,sp)
| RawAbsExpr (_,_,sp) | RawBranchExpr (_,sp)
| RawSeqExpr (_,_,sp) | RawFailExpr (_,sp)
| RawSuccessExpr sp | RawEconExpr (_,_,sp)
| RawFconExpr (_,_,sp) | RawFreshExpr (_,sp)
| RawDefnExpr (_,sp) | RawLetRecExpr(_,_,_,_,_,sp)
| RawUnbindExpr (_,_,_,_,_,_,_,sp) | RawDistinctExpr (_,sp)
| RawExistsExpr (_,_,_,sp) | RawTreeExpr (_,sp) | RawPrintExpr (_,sp)-> sp
let getSpanAtomic = function (_,_,sp) -> sp

(* ***************************************************** *)
(*                                                       *)
(*   EXPAND A RAW EXPRESSION BY TRANSLATING OUT ANY      *)
(*   OF THE DEFINED FEATURES, PRIOR TO TYPECHECKING,     *)
(*   COMPILATION AND EXECUTION.                          *)
(*                                                       *)
(*   This could be done within the semantic actions of   *)
(*   the parser, but this would require the arguments    *)
(*   "dec" and "rdec" to be passed as arguments to       *)
(*   Ocamlyacc. I'm not sure that this is even possible. *)
(*                                                       *)
(* ***************************************************** *)

(* Dummy variables used within the translation. *)
let rawDummyX : rawId = ("%x",dummy_span)        (* % is forbidden in concrete syntax! *)
let rawDummyRelFn : rawId = ("%rel", dummy_span) (* % is forbidden in concrete syntax! *)

(* Compute the free names of a raw tree. Returns strings. *)
(* let treeFN g = *)
(*   let rec f = function *)
(*       RawNameTree x -> [(getStr x)] *)
(*     | RawUnitTree _ -> [] *)
(*     | RawAbsTree (x,_,g,_) -> List.filter (fun y -> (getStr x) <> y) (f g) *)
(*     | RawDataTree (_,g,_) -> f g *)
(*     | RawTupTree (gs,_) -> List.fold_left (fun xs g -> merge xs (f g)) [] gs *)
(*   in *)
(*   f g *)

(* Compute the free names of a raw tree. Returns rawIds. *)
let treeFN g =
  let rec f = function
      RawNameTree x -> [x]
    | RawUnitTree _ -> []
    | RawAbsTree (x,_,g,_) -> List.filter (fun y -> not(cmpId x y)) (f g)
    | RawDataTree (_,g,_) -> f g
    | RawTupTree (gs,_) -> List.fold_left (fun xs g -> merge2 cmpId xs (f g)) [] gs
  in
  f g

(* Expand out a raw tree with the relevant freshnesses etc. *)
let expandRawTree g =
  (* Counter for creating new variables. *)
  let ctr = ref 0 in
  let nextVar () =
    let i = !ctr in
    begin
      ctr := !ctr + 1;
      "%" ^ (string_of_int i)
    end
  in
  (* Do recursively-defined translation. *)
  let rec doTrn (env:(string * string) list) (nenv:(rawId * rawType) list) g = match g with
      RawNameTree a ->
	begin
	  try
	    let x = List.assoc (getStr a) env in
	    (nenv, RawVarExpr(x,(getSpanId a)))
	  with
	    Not_found -> failwith "expandRawTree - name not found!"
	end
    | RawUnitTree sp -> (nenv, RawUnitExpr sp)
    | RawDataTree (k,g,sp) ->
	let (nenv',v) = doTrn env nenv g in (nenv', RawDataExpr(k,v,sp))
    | RawTupTree (gs,sp) ->
        let (nenv',vs) = List.fold_left
	  (fun (nenv_i,vs_i) g -> let (nenv_j,v_j) = doTrn env nenv_i g in (nenv_j,v_j::vs_i))
	  (nenv,[]) gs
	in
	(nenv', RawTupExpr((List.rev vs),sp))
    | RawAbsTree (a,t,g,sp) ->
	let x = nextVar () in
        let x_id = (x,(getSpanId a)) in
        let (vars',v) = doTrn (((getStr a),x)::env) ((x_id,t)::nenv) g in
        (vars', RawAbsExpr((RawVarExpr x_id),v,sp))
  in
  (* Use the definitions to build the translated expression. *)
  let free_xs = treeFN g in
  let initial_envt = List.map (fun x -> let xstr = getStr x in (xstr,xstr)) free_xs in
  let nenv',v_g = doTrn initial_envt [] g in
  let freshfor_xs = remDup(free_xs@(List.map fst nenv')) in
  let freshfor_exprs = List.map (fun x -> RawVarExpr x) freshfor_xs in
  let base_expr =
    RawSeqExpr(RawDistinctExpr(freshfor_exprs,dummy_span),v_g,dummy_span)
  in
  List.fold_right (fun (x,t) e ->
    RawLetExpr(x,[],(RawSomeExpr(t,dummy_span)),e,dummy_span))
    nenv' base_expr

(* Translate a rawPat into a rawExpr. *)
let rec trnRawPat = function
  RawVarPat x -> RawVarExpr x
| RawUnitPat sp -> RawUnitExpr sp
| RawAbsPat(x,rp,sp) -> RawAbsExpr(RawVarExpr x, trnRawPat rp, sp)
| RawTupPat(rps,sp) -> RawTupExpr(List.map trnRawPat rps, sp)
| RawDataPat(k,rp,sp) -> RawDataExpr(k, trnRawPat rp, sp)

(* Expand out schematic inductive definitions. *)
let rec expandRawDefn (dec:dtyDec) (rdec:relDec) (rls:rawDefn) =
  (* Expand an atomic formula. *)
  let expandRawAtomic (r,(rps,sp'),sp) =
    match rps with
	[] -> failwith "expandRawAtomic - empty list of expressions"
      | [rp] -> RawDataExpr(r,(trnRawPat rp),sp)
      | rps -> RawDataExpr(r,RawTupExpr((List.map trnRawPat rps),sp'),sp)
  in
  (* Expand a single schematic rule. *)
  let expandRule ((prem,conc,name,vartys,sp):rawRule) =
    let eqre = RawEconExpr(RawVarExpr(rawDummyX),(expandRawAtomic conc),dummy_span) in
    let body = RawSeqExpr(eqre,(expandPremise dec rdec prem),dummy_span) in
    let trnexpr = List.fold_left
      (fun re (x,ty) -> RawLetExpr(x,[],RawSomeExpr(ty,dummy_span),re,dummy_span))
      body vartys
    in
    let namestr = begin match name with Some n -> n | None -> "_" end in
    RawSeqExpr(RawPrintExpr(("DEBUG: trying rule: " ^ namestr),dummy_span),trnexpr,dummy_span)
  in
  let branches = RawBranchExpr((List.map expandRule rls),dummy_span) in
  RawRecExpr(rawDummyRelFn,rawDummyX,RawRelTy,RawPropTy,branches,dummy_span)

(* Expand any "relation" data terms (in a rule premise) into recursive calls. *)
and expandPremise (dec:dtyDec) (rdec:relDec) (re:rawExpr) =
  let fix = (expandPremise dec rdec) in
  match re with
    RawTupExpr (es,sp) -> RawTupExpr((List.map fix es),sp)
  | RawAbsExpr (e,e',sp) -> RawAbsExpr((fix e),(fix e'),sp)
  | RawDataExpr (k,e,sp) -> let e' = fix e in
      begin match findDtyCon dec (getStr k) with
        Some _ -> RawDataExpr(k,e',sp)
      | None ->
	  begin match findRelCon rdec (getStr k) with
	      Some _ ->
		RawAppExpr([RawVarExpr rawDummyRelFn; (RawDataExpr(k,e',sp))],dummy_span)
	    | None -> raise (UnknownCon k)
	  end
      end
  | RawProjExpr (e,i,sp) -> RawProjExpr((fix e),i,sp)
  | RawLamExpr (x,ty,e,sp) -> RawLamExpr(x,ty,(fix e),sp)
  | RawRecExpr (f,x,inty,outty,e,sp) -> RawRecExpr(f,x,inty,outty,(fix e),sp)
  | RawEconExpr (e,e',sp) -> RawEconExpr((fix e),(fix e'),sp)
  | RawFconExpr (e,e',sp) -> RawFconExpr((fix e),(fix e'),sp)
  | RawBranchExpr (es,sp) -> RawBranchExpr((List.map fix es),sp)
  | RawSeqExpr (e,e',sp) -> RawSeqExpr((fix e),(fix e'),sp)
  | RawCaseExpr (e,kxes,sp) ->
      let bodies = List.map (fun (k,x,e) -> (k,x,(fix e))) kxes in
	RawCaseExpr((fix e),bodies,sp)
  | RawAppExpr (es,sp) -> RawAppExpr((List.map fix es),sp)
  | RawDistinctExpr (es,sp) -> RawDistinctExpr((List.map fix es),sp)
  | RawLetExpr(x,xts,e,e',sp) ->
      expand dec rdec (RawLetExpr(x,xts,(fix e),(fix e'),sp))
  | RawLetRecExpr(f,xts,outty,e,e',sp) ->
      expand dec rdec (RawLetRecExpr(f,xts,outty,(fix e),(fix e'),sp))
  | RawUnbindExpr(e,x,x',nty,ety,e',use_fresh,sp) ->
      expand dec rdec (RawUnbindExpr((fix e),x,x',nty,ety,(fix e'),use_fresh,sp))
  | RawExistsExpr(x,t,e,sp) ->
      expand dec rdec (RawExistsExpr(x,t,(fix e),sp))
  | (RawDefnExpr _) as exp -> expand dec rdec exp (* ?????? *)
  | (RawTreeExpr _) as exp -> expand dec rdec exp (* ?????? *)
  | e -> e

(* Translate away any defined constructs within a raw expression! *)
and expand (dec:dtyDec) (rdec:relDec) (re:rawExpr) =
  let elim = (expand dec rdec) in
  match re with
      RawTupExpr (es,sp) -> RawTupExpr((List.map elim es),sp)
    | RawAbsExpr (e,e',sp) -> RawAbsExpr((elim e),(elim e'),sp)
    | RawDataExpr (k,e,sp) -> RawDataExpr(k,(elim e),sp)
    | RawProjExpr (e,i,sp) -> RawProjExpr((elim e),i,sp)
    | RawLamExpr (x,ty,e,sp) -> RawLamExpr(x,ty,(elim e),sp)
    | RawRecExpr (f,x,inty,outty,e,sp) -> RawRecExpr(f,x,inty,outty,(elim e),sp)
    | RawEconExpr (e,e',sp) -> RawEconExpr((elim e),(elim e'),sp)
    | RawFconExpr (e,e',sp) -> RawFconExpr((elim e),(elim e'),sp)
    | RawBranchExpr (es,sp) -> RawBranchExpr((List.map elim es),sp)
    | RawSeqExpr (e,e',sp) -> RawSeqExpr((elim e),(elim e'),sp)
    | RawCaseExpr (e,kxes,sp) ->
	let bodies = List.map (fun (k,x,e) -> (k,x,(elim e))) kxes in
	  RawCaseExpr((elim e),bodies,sp)
    | RawAppExpr (es,sp) -> RawAppExpr((List.map elim es),sp)
    | RawDistinctExpr (es,sp) -> RawDistinctExpr((List.map elim es),sp)
    | RawDefnExpr (d,sp) -> expandRawDefn dec rdec d
    | RawLetExpr (x,xts,e,e',sp) ->
	let e_fixed = List.fold_right
	  (fun (x,t) e_acc -> RawLamExpr(x,t,e_acc,dummy_span))
	  xts (elim e)
	in
	RawLetExpr(x,[],e_fixed,(elim e'),sp)
    | RawLetRecExpr(f,xts,outty,e,e',sp) ->
	begin
	  match xts with
	      [] -> failwith "expand - no arguments for function in let rec"
	    | (x,t)::xts' ->
		let functy = List.fold_right
		  (fun t ty -> RawFunTy(t,ty))
		  (List.map snd xts') outty
		in
		let funcbody = List.fold_right
		  (fun (x,t) e_acc -> RawLamExpr(x,t,e_acc,dummy_span))
		  xts' (elim e)
		in
		let func = RawRecExpr(f,x,t,functy,funcbody,dummy_span) in
		  RawLetExpr(f,[],func,(elim e'),sp)
	end
    | RawUnbindExpr(e,x,x',nty,ety,e',use_fresh,sp) ->
        let mk_nty = if use_fresh then
	  RawFreshExpr(nty, dummy_span)
	else
	  RawSomeExpr(RawTypeId (getStr nty), dummy_span)
	in
	let mk_ety = RawSomeExpr(ety, dummy_span) in
	let abs_pat = RawAbsExpr(RawVarExpr x,RawVarExpr x',dummy_span) in
	RawLetExpr(x,[],mk_nty,RawLetExpr(x',[],mk_ety,
          RawSeqExpr(RawEconExpr((elim e),abs_pat,dummy_span),(elim e'),dummy_span),dummy_span),sp)
    | RawExistsExpr(x,t,e,sp) ->
        RawLetExpr(x,[],RawSomeExpr(t,dummy_span),(elim e),sp)
    | RawTreeExpr(g,sp) -> expandRawTree g
    | e -> e

(* ********************************************************* *)
(*                                                           *)
(*   COMPILE A RAW AST TO A LIST OF MACHINE INSTRUCTIONS.    *)
(*   USE A LIST OF STRING VARIABLES AS THE SYMBOL TABLE.     *)
(*                                                           *)
(* ********************************************************* *)

(* Type for symbol tables. *)
type symtab = string list

(* Compile an expression into a list of machine instructions. *)
let compile dec ndec (st:symtab) re =
  let compile_error str = failwith ("Compilation error - " ^ str) in
  let classify = classifyIds ndec (List.map fst dec) in
  let rec cmp st = function
    RawVarExpr (x,sp) -> let i = unoption(index x st) in [IVar i]
  | RawEvarExpr (u,sp) -> let i = int_of_xyzstring u in [IEvar i]
  | RawDataExpr (k,e,sp) -> (cmp st e) @ [IData (getStr k)]
  | RawTupExpr (es,sp) ->
      (linkMap (cmp st) es) @ [ITuple (List.length es)]
  | RawAbsExpr (e,e',sp) -> (cmp st e') @ (cmp st e) @ [IAbs]
  | RawUnitExpr sp -> [IUnit]
  | RawSuccessExpr sp -> [IYes]
  | RawFailExpr (t,sp) -> [IFail]
  | RawProjExpr (e,i,sp) -> (cmp st e) @ [IProj (getInt i)]
  | RawLamExpr (x,ty,e,sp) ->
      [IClosure((cmp ((getStr x)::st) e) @ [IPopEnv])]
  | RawRecExpr (f,x,inty,outty,e,sp) ->
      [IRecClosure((cmp ((getStr x)::(getStr f)::st) e) @ [IPopEnv])]
  | RawEconExpr (e,e',sp) -> (cmp st e') @ (cmp st e) @ [IEcon;IYes]
  | RawFconExpr (e,e',sp) -> (cmp st e') @ (cmp st e) @ [IFcon;IYes]
  | RawBranchExpr (es,sp) -> [IBranch(List.map (cmp st) es)]
  | RawSomeExpr (t,sp) ->
      let trnt = trnRawType classify t in [IExists trnt]
  | RawFreshExpr (n,sp) -> [IFresh (getStr n)]
  | RawSeqExpr (e,e',sp) -> (cmp st e) @ [IDiscardStkTop] @ (cmp st e')
  | RawCaseExpr (e,kxes,sp) -> (cmp st e) @ [ICase (List.map
      (fun (k,x,e) -> ((getStr k),((cmp ((getStr x)::st) e)@[IDiscardEnvTop]))) kxes)]
  | RawLetExpr (x,xts,e,e',sp) ->
      begin
	match xts with
	  [] -> (cmp st e) @ [IStkToEnv] @ (cmp ((getStr x)::st) e') @ [IDiscardEnvTop]
	| _ -> compile_error "non-trivial let expressions should have been expanded"
      end
  | RawAppExpr (es,sp) ->
      begin
        match es with
          [] -> compile_error "empty application node"
        | [e] -> compile_error "singleton application node"
        | (e1::e2::es) ->
            let rec mkExtras = function
              [] -> []
            | e::es -> (cmp st e) @ [ICall] @ (mkExtras es)
            in
            (cmp st e1) @ (cmp st e2) @ [ICall] @ (mkExtras es)
      end
  | RawDistinctExpr (es,sp) ->
      (linkMap (cmp st) es) @ [IDistinct (List.length es);IYes]
  | RawPrintExpr (str,sp) -> [IPrint str]	
  | RawDefnExpr _ -> compile_error "inductive definitions should have been expanded"
  | RawLetRecExpr _ -> compile_error "let rec expressions should have been expanded"
  | RawUnbindExpr _ -> compile_error "unbind expressions should have been expanded"
  | RawExistsExpr _ -> compile_error "exists expressions should have been expanded"
  | RawTreeExpr _ -> compile_error "tree expressions should have been expanded"
  in
  cmp st re

(* *********************************************************************************** *)

(* Produce a string representation of a raw expression, for debugging purposes. *)
let rec string_of_rawExpr re =
  let paren str = "(" ^ str ^ ")" in
  let sot = soRawType in
  let soe = string_of_rawExpr in
  let error str = failwith ("string_of_rawExpr - " ^ str) in
  match re with
      RawVarExpr (x,sp) -> x
    | RawEvarExpr (u,sp) -> "?" ^ u
    | RawDataExpr (k,e,sp) -> "(" ^ (getStr k) ^ " " ^ (soe e) ^ ")"
    | RawTupExpr (es,sp) -> string_of_enclist "(TUPBEGIN(" ")TUPEND)" soe ", " es
    | RawAbsExpr (e,e',sp) -> "(<" ^ (soe e) ^ ">" ^ (soe e') ^")"
    | RawUnitExpr sp -> "()"
    | RawSuccessExpr sp -> "yes"
    | RawFailExpr (t,sp) -> "no" ^ (sot t)
    | RawProjExpr (e,i,sp) -> (soe e) ^ "." ^ (string_of_int (getInt i))
    | RawLamExpr (x,ty,e,sp) ->
	"(fn " ^ (paren (getStr x) ^ ":" ^ (sot ty)) ^ " -> " ^ (soe
	  e) ^ ")"
    | RawRecExpr (f,x,inty,outty,e,sp) ->
	"(fun " ^ (getStr f) ^ (paren((getStr x) ^ ":" ^ (sot inty))) ^
	  ":" ^ (sot outty) ^ " = " ^ (soe e) ^ ")"
    | RawEconExpr (e,e',sp) -> (soe e) ^ " = " ^ (soe e')
    | RawFconExpr (e,e',sp) -> (soe e) ^ " # " ^ (soe e')
    | RawBranchExpr (es,sp) -> "(" ^ (string_of_list soe " || " es) ^ ")"
    | RawSomeExpr (t,sp) -> "(some " ^ (sot t) ^ ")"
    | RawFreshExpr (n,sp) -> "(fresh " ^ (getStr n) ^ ")"
    | RawSeqExpr (e,e',sp) -> "(" ^ (soe e) ^ " & " ^ (soe e') ^ ")"
    | RawCaseExpr (e,kxes,sp) ->
        "(case " ^ (soe e) ^ " of " ^
	  (string_of_list (fun (k,x,e) -> (getStr k) ^ " " ^ (getStr x) ^ " -> " ^ (soe e))
	     " | " kxes) ^ ")"
    | RawLetExpr (x,xts,e,e',sp) ->
	begin
	  match xts with
	      [] -> "(let " ^ (getStr x) ^ " = " ^ (soe e) ^ " in " ^ (soe e') ^ ")"
	    | _ -> error "non-trivial let expressions should have been expanded"
	end
    | RawAppExpr (es,sp) -> "(" ^ (string_of_list soe " " es) ^ ")"
    | RawDistinctExpr (es,sp) -> "(distinct" ^ (string_of_enclist "("
	")" soe ", " es) ^ ")"
    | RawPrintExpr (str,sp) -> "(print \"" ^ str ^ "\")"
    | RawDefnExpr _ -> error "inductive definitions should have been expanded"
    | RawLetRecExpr _ -> error "let rec expressions should have been expanded"
    | RawUnbindExpr _ -> error "unbind expressions should have been expanded"
    | RawExistsExpr _ -> error "exists expressions should have been expanded"
    | RawTreeExpr _ -> error "tree expressions should have been expanded"
