open Globals
open Library
open Location
open Machine
open Types
open Errors

(* Raw identifiers *)
type rawId = string * span

(* Convert a rawId into a normal string *)
val getStr : rawId -> string

(* Raw integer expression, for projection *)
type rawInt = int * span

(* Convert rawInt into a normal int *)
val getInt : rawInt -> int

(* A string used for implementing wildcard "patterns". *)
val wildcard : string

(* Raw alphaML types. *)
type rawType = 
  RawUnitTy
| RawPropTy
| RawRelTy
| RawTypeId of string
| RawTupTy  of rawType list
| RawFunTy  of rawType * rawType
| RawAbsTy  of string * rawType

(* to_string function for raw types *)
val soRawType : rawType -> string

(* Classify type identifiers. *)
val classifyIds : string list -> string list -> string -> ty

(* Translate a rawType into an ty, using the supplied
 * function to interpret any type identifiers. *)
val trnRawType : (string -> ty) -> rawType -> ty

(* Raw datatype declaration, as returned from parser *)
type rawDtyDec = (string * ((string * rawType) list)) list

(* to_string function for raw datatype declarations *)
val rawdec_toString : rawDtyDec -> string

(* Raw relation declaration, as returned from parser *)
type rawRelDec = (string * rawType) list

(* to_string function for raw relation declarations *)
val rawRelDec_toString : rawRelDec -> string

(* Raw ground trees. *)
type rawTree =
  RawNameTree of rawId
| RawUnitTree of span
| RawAbsTree of rawId * rawType * rawTree * span
| RawDataTree of rawId * rawTree * span
| RawTupTree of (rawTree list) * span

(* Raw expressions *)
type rawExpr =
  RawVarExpr of rawId
| RawEvarExpr of rawId
| RawLamExpr of rawId * rawType * rawExpr * span
| RawRecExpr of rawId * rawId * rawType * rawType * rawExpr * span
| RawAppExpr of (rawExpr list) * span
| RawProjExpr of rawExpr * rawInt * span
| RawCaseExpr of rawExpr * ((rawId * rawId * rawExpr) list) * span
| RawDataExpr of rawId * rawExpr * span
| RawTupExpr of (rawExpr list) * span
| RawUnitExpr of span
| RawSomeExpr of rawType * span
| RawAbsExpr of rawExpr * rawExpr * span
| RawBranchExpr of (rawExpr list) * span
| RawSeqExpr of rawExpr * rawExpr * span
| RawFailExpr of rawType * span
| RawSuccessExpr of span
| RawEconExpr of rawExpr * rawExpr * span
| RawFconExpr of rawExpr * rawExpr * span
| RawFreshExpr of rawId * span
| RawDistinctExpr of (rawExpr list) * span
| RawDefnExpr of rawDefn * span
| RawLetExpr of rawId * (rawId * rawType) list * rawExpr * rawExpr * span
| RawLetRecExpr of rawId * (rawId * rawType) list * rawType * rawExpr * rawExpr * span
| RawUnbindExpr of rawExpr * rawId * rawId * rawId * rawType * rawExpr * bool * span
| RawExistsExpr of rawId * rawType * rawExpr * span
| RawTreeExpr of rawTree * span
| RawPrintExpr of string * span
(* Raw patterns *)
and rawPat =
  RawVarPat of rawId  
| RawUnitPat of span
| RawAbsPat of rawId * rawPat * span
| RawTupPat of (rawPat list) * span
| RawDataPat of rawId * rawPat * span
(* Raw rules and definitions *)
and rawAtomic = rawId * ((rawPat list) * span) * span
and rawRule =
  rawExpr * rawAtomic * (string option) * ((rawId * rawType) list) * span
and rawDefn = rawRule list

(* Get span information out of ASTs *)
val getSpanId : rawId -> span
val getSpanInt : rawInt -> span
val getSpanExpr : rawExpr -> span
val getSpanAtomic : rawAtomic -> span

(* Expand any defined constructs within a raw expression! *)
val expand : dtyDec -> relDec -> rawExpr -> rawExpr

(* Type for symbol tables. *)
type symtab = string list

(* Compile an expression into a list of machine instructions. *)
val compile : dtyDec -> ntyDec -> symtab -> rawExpr -> instr list

(* Produce a string representation of a raw expression.
 * For debugging the translation of raw inductive definitions! *)
val string_of_rawExpr : rawExpr -> string
