open Location
open Types

(* Exceptions for lexical and syntactic errors *)
exception SyntaxError of span
exception IllegalToken of span
exception CommentNotTerminated of span

(* Fail with an error that the type checker should prevent. *)
let failTC str = failwith
  ("*** TYPE CHECKER SHOULD PREVENT THIS ERROR! " ^
   "(reference " ^ str ^ ") ***")

(* Exception indicating a runtime error *)
exception Machine_error of string

(* Exception raised if constraint solver needs to backtrack. *)
exception ConstraintBacktrack
(* Exception raised if constraint solver detects satisfiability. *)
exception ConstraintSuccess

(* EXCEPTIONS THROWN WHEN TYPECHECKING FAILS. *)

(* Some error messages may involve an actual type or a textual description. *)
type tyerr = TYPE of ty | DESCR of string
let soTyerr = function TYPE t -> soType t | DESCR x -> x

(* If an unexpected type is seen. First type is what was seen,
 * second type is what was expected. *)
exception WrongType of span * ty * tyerr
(* If a variable is not bound. *)
exception UnboundVar of (string * span)
(* If projection index is out of range. *)
exception ProjRange of (int * span) * int
(* If a case expression isn't properly formatted. *)
exception CaseError of span * string
(* If a constructor is unknown. *)
exception UnknownCon of (string * span)
(* If we don't see an equality type when we should. *)
exception NotEqTy of span * ty
(* If we don't see an inhabited type when we should. *)
exception NotInhab of span * ty
(* Unbound existential variable *)
exception UnboundEvar of (string * span)
(* If we were expecting a nametype. *)
exception NotNty of (string * span)

(* Exceptions raised when a datatype, nametype, relation
 * or constructor has already been seen. *)
exception AlreadyDty of string
exception AlreadyNty of string
exception AlreadyCon of string
exception AlreadyRel of string

(* Exception raised for a type identifier that's already bound *)
exception UnboundTypeId of string

(* Exception raised if the user tries to abstract over
 * something that isn't a nametype. *)
exception AbsError of string

(* Exception raised for an unknown interpreter directive. *)
exception UnknownDirective of string
