open Library
open Errors
open Machine
open Solver
open Types
open Constraint

(* Abstract machine runtime. *)

(* Report a runtime error *)
let error msg = raise (Machine_error msg)

(* The state of the machine consists of a frame,
 * a stack of machine values, a list of environments,
 * a Vardata.t which records the types of generated variables
 * and their substitutions, and a Solver.prob which is the
 * leftover constraints after solving! *)
type state = frame * stack * (environment list) * Vardata.t * prob

(* Environment lookup helper function. *)
let lookup i envs =
  match envs with
    env::_ -> List.nth env i
  | [] -> error "lookup"

(* Split the stack into top n and rest. *)
let split stk n =
  let rec do_split top rest n =
    match (n,rest) with
      0,_ -> ((List.rev top),rest)
    | n,(mv::rest) -> do_split (mv::top) rest (n-1)
    | _ -> error "split"
  in
  do_split [] stk n

(* Pop the stack elements needed to execute an application,
 * i.e. a value and closure. *)
let pop_app = function
  mv :: MClosure(f,e) :: stk -> (f, e, mv, stk)
| _ -> error "pop_app"

(* Cons a machine value onto the top environment on the environment stack. *)
let envcons mv = function
  [] -> [[mv]]
| (env::envs) -> ((mv::env)::envs)

(* "exec instr frms stk envs vd cs dec" executes instruction [instr] in the
 * given state "(frms, stk, envs, vd, cs)".
 * The result is a list of new states.
 * The dec argument gives types for data constructors (needed by the
 * constraint solver to do narrowing). *)
let exec instr frm stk envs vd cs dec = match instr with
  IVar i -> [frm, ((lookup i envs)::stk), envs, vd, cs]
| IEvar i -> [frm, ((MEvar i)::stk), envs, vd, cs]
| IUnit -> [frm, (MUnit::stk), envs, vd, cs]
| IYes -> [frm, (MYes::stk), envs, vd, cs]
| IData k ->
    begin
      match stk with
        [] -> error "IData" (* ??? *)
      | mv::stk -> [frm, ((MData (k,mv))::stk), envs, vd, cs]
    end
| ITuple n ->
    let mvs,stk = split stk n in
    [frm, ((MTuple (List.rev mvs))::stk), envs, vd, cs]
| IAbs ->
    begin
      match stk with
        mv::mv'::stk -> [frm, ((MAbs (mv,mv'))::stk), envs, vd, cs]
      | _ -> error "IAbs"
    end
| IClosure f ->
    begin
      match envs with
        env::_ ->
          let c = MClosure (f,env) in
          [frm, (c::stk), envs, vd, cs]
      | _ -> error "IClosure"
    end
| IRecClosure f ->
    begin
      match envs with
        env::_ ->
          let rec c = MClosure (f,(c::env)) in
          [frm, (c::stk), envs, vd, cs]
      | _ -> error "IRecClosure"
    end
| IFail ->
    begin
      debug "Execute IFail instruction.";
      []
    end
| IBranch fs ->
    List.map (fun f -> ((f@frm), stk, envs, vd, cs)) fs
| ICall ->
    let (f, e, mv, stk) = pop_app stk in
    [f@frm, stk, ((mv::e)::envs), vd, cs]
| IPopEnv ->
    begin
      match envs with
        _::envs -> [frm, stk, envs, vd, cs]
      | [] -> error "IPopEnv"
    end
| IDiscardStkTop ->
    begin
      match stk with
        _::stk -> [frm, stk, envs, vd, cs]
      | _ -> error "IDiscardStkTop"
    end
| IDiscardEnvTop ->
    begin
      match envs with
        (_::env)::envs -> [frm, stk, (env::envs), vd, cs]
      | _ -> error "IDiscardEnvTop"
    end
| IStkToEnv ->
    begin
      match stk with
        (mv::stk) -> [frm, stk, (envcons mv envs), vd, cs]
      | [] -> error "IStkToEnv"
    end
| IExists ty ->
    let (i,vd') = Vardata.addEvar vd ty in
    [frm, ((MEvar i)::stk), envs, vd', cs]
| IProj i ->
    begin
      match stk with
        (MTuple mvs)::stk ->
          let mv = List.nth mvs (i-1) in
          [frm, (mv::stk), envs, vd, cs]
      | (MEvar x)::stk ->
          begin
	    match (Vardata.getSubst vd x) with
		Some(MTuple mvs) ->
		  let mv = List.nth mvs (i-1) in
   	          [frm, (mv::stk), envs, vd, cs]
	      | Some _ -> error "IProj - MEvar - wrong subst"
	      | None ->
		  begin
		    match (Vardata.getType vd x) with
			TupTy etys ->
			  let j = Vardata.lastEvar vd in
			  let mkevars = List.map (fun ety -> IExists ety) etys in
			  let n = List.length etys in
			  let f' = mkevars @ [ITuple n; IEcon; IEvar (j+i)] in
			    [(f'@frm), ((MEvar x)::stk), envs, vd, cs]
		      | _ -> error "IProj - MEvar - wrong type"
		  end
	  end
      | _ -> error "IProj"
    end
| ICase kfs ->
    begin
      let rec find_match k = function
          [] -> error "ICase - MData"
        | (k',f)::kfs -> if k=k' then f else find_match k kfs
      in
      match stk with
        (MData (k,mv))::stk ->
          let f = find_match k kfs in
          [(f@frm), stk, (envcons mv envs), vd, cs]
      | (MEvar x)::stk ->
	  begin
	    match (Vardata.getSubst vd x) with
		Some(MData(k,mv)) ->
		  let f = find_match k kfs in
		  [(f@frm), stk, (envcons mv envs), vd, cs]
	      | Some _ -> error "ICase - MEvar - wrong subst"
	      | None ->
		  let makeBranch = fun (k,f) ->
		    begin
		      match (Types.findDtyCon dec k) with
			  Some (ety,_) -> [IExists ety; IData k; IEcon] @ f
			| None -> error "ICase - MEvar - unknown constructor"
		    end
		  in
		  let j = Vardata.lastEvar vd in
		    List.map
		      (fun kf -> (((makeBranch kf)@frm), ((MEvar x)::stk),
				  (envcons (MEvar (j+1)) envs), vd, cs))
		      kfs	    
	  end
      | _ -> error "ICase"
    end
| IEcon ->
    begin
      match stk with
        mv::mv'::stk ->
          begin
	    debug "Adding equality constraint...";
            match (Solver.addConstraints dec vd cs [mkEq mv mv']) with
 		Some (vd',cs') ->
		begin
		  debug "Leaving solver!";
		  [frm, stk, envs, vd', cs']
		end
              | None -> []
          end
      | _ -> error "IEcon"
    end
| IFcon ->
    begin
      match stk with
        (MEvar i)::mv'::stk ->
          begin
	    debug "Adding freshness constraint...";
            match (Solver.addConstraints dec vd cs [mkFr i mv']) with
              Some (vd',cs') ->
		begin
		  debug "Leaving solver!";
		  [frm, stk, envs, vd', cs']
		end
              | None -> []
          end
      | _ -> error "IFcon"
    end
| IDistinct n ->
    begin
      let mvs,stk = split stk n in
      let mvs = List.rev mvs in
      let rec mkFcons = function
	  [] -> []
	| (MEvar i)::mvs -> (List.map (fun mv -> mkFr i mv) mvs) @ mkFcons mvs
	| _ -> error "IDistinct"
      in
      let ncs = mkFcons mvs in
      begin
	debug "Adding distinctness constraints...";
	match (Solver.addConstraints dec vd cs ncs) with
	    Some (vd',cs') ->
	      begin
		debug "Leaving solver!";
		[frm,stk,envs,vd',cs']
	      end
	  | None -> []
      end
    end
| IFresh n ->
    begin
      let (i,vd') = Vardata.addEvar vd (Nty n) in
      let rec mkFcons = function
	  0 -> [mkFr i (MEvar 0)]
	| n -> (mkFr i (MEvar n))::(mkFcons (n-1))
      in
      begin
	debug "Adding constraints from a fresh...";
	match (Solver.addConstraints dec vd' cs (mkFcons (i-1))) with
	    Some (vd'',cs'') ->
	      begin
		debug "Leaving solver!";
		[frm, ((MEvar i)::stk), envs, vd'', cs'']
	      end
	  | None -> error "IFresh"
      end
    end
| IPrint str ->
    begin
      debug str;
      [frm, (MUnit::stk), envs, vd, cs]
    end

(* Default TTL value for breadth-first searching. *)
let initial_ttl = ref 50

(* Run the machine in breadth-first mode. *)
let run_breadth dec ss =
  let rec do_loop = function
    (* Current computation has terminated successfully! *)
      (_,([],[mv],_,vd,cs),ss) ->
	begin
	  debug "Successful termination.";
	  Some(mv,vd,cs,ss)
	end
    (* Run out of TTL for current computation. *)
    | (0,s,ss) ->
        begin match ss with
          (* Keep going if nothing else to do! *)
            [] ->
	      begin
		debug "Resuming only computation.";
		do_loop ((!initial_ttl),s,[])
	      end
          (* Move onto the next computation! *)
          | s'::ss' ->
	      begin
		debug "Context switching.";
		do_loop ((!initial_ttl),s',(ss'@[s]))
	      end
        end
    (* Keep going on current computation! *)
    | (n,((i::frm),stk,envs,vd,cs),ss) ->
        begin
	  debug ("Threads waiting: " ^ (string_of_int (List.length ss)));
	  match (exec i frm stk envs vd cs dec) with
          [] -> begin
	    debug "Computation step failed.";
	    match ss with
                [] ->
		  begin
		    debug "Finite failure.";
		    None
		  end	
              | s'::ss' ->
		  begin
		    debug "Continue with next computation.";
		    do_loop ((!initial_ttl),s',ss')
		  end
          end
        | s'::ss' ->
	    do_loop ((n-1),s',(ss@ss')) (* NB: BREADTH FIRST! *)
        end
    (* Something else means something's wrong! *)
    | _ -> error "invalid machine state"
  in
  match ss with
    [] -> failwith "Nothing to execute!"
  | s::ss -> 
      begin
	debug "Starting breadth-first search.";
	do_loop ((!initial_ttl),s,ss)
      end

(* Run the machine in depth-first mode. *)
let run_depth dec ss =
  let rec do_loop = function
    (* Current computation has terminated successfully! *)
      (([],[mv],_,vd,cs),ss) -> Some(mv,vd,cs,ss)
    (* Keep going on current computation! *)
    | (((i::frm),stk,envs,vd,cs),ss) ->
        begin match (exec i frm stk envs vd cs dec) with
          [] -> begin match ss with
                  [] -> None
                | s'::ss' -> do_loop (s',ss')
                end
        | s'::ss' ->
	    begin
	      debug "Starting depth-first search.";
	      do_loop (s',(ss'@ss)) (* NB: DEPTH FIRST! *)
	    end
        end
    (* Something else means something's wrong! *)
    | _ -> error "invalid machine state"
  in
  match ss with
    [] -> failwith "Nothing to execute!"
  | s::ss -> do_loop (s,ss)

(* Return a string representation of abstract machine state. *)
let soState ((frm,stk,envs,vd,cs):state) = 
  "{*\n" ^ (soFrame frm) ^ "\n" ^ (soEnvs envs) ^ "\n*}" (* ?????????? *)
