(* Lexical analyser for the alphaML language *)

{
 open Lexing
 open Parser
 open Location
 open Globals
 open Errors

 (* Memory-efficient way of using a reasonable number of keyword tokens *)
 let keyword_table = Hashtbl.create 72
 let _ = List.iter (fun (kwd, tok) -> Hashtbl.add keyword_table kwd tok)
     [ ("let", LET);
       ("in", IN);
       ("case", CASE);
       ("of", OF);
       ("fn", FN);
       ("fun", FUN);
       ("unit", UNIT);
       ("nametype", NAMETYPE);
       ("datatype", DATATYPE);
       ("relation", RELATION);
       ("and", AND);
       ("some", SOME);
       ("no", FAIL);
       ("yes", SUCCEED);
       ("fresh", FRESH);
       ("where", WHERE);
       ("on", ON);
       ("off", OFF);
       ("rec", REC);
       ("unbind",UNBIND);
       ("unbind_fresh",UNBIND_FRESH);
       ("as",AS);
       ("distinct",DISTINCT);
       ("exists",EXISTS)
     ]

 (* Increment the line count and start of line offset *)
 let inc_lines lexbuf =
   let new_curr_p =
     { pos_fname = lexbuf.lex_curr_p.pos_fname;
       pos_lnum = lexbuf.lex_curr_p.pos_lnum + 1;
       pos_bol = lexbuf.lex_curr_p.pos_cnum;
       pos_cnum = lexbuf.lex_curr_p.pos_cnum }
   in
   lexbuf.lex_curr_p <- new_curr_p

  (* Modify state on the location of the last seen lexeme -
     for giving syntax error exceptions *)
  let update_lexeme lexbuf =
    curr_lexeme_start := loc_from_pos (lexeme_start_p lexbuf);
    curr_lexeme_finish := loc_from_pos (lexeme_end_p lexbuf)    

  (* Code to "get" the current span - used a few times! *)
  let get_span lexbuf = span_from_poss (lexeme_start_p lexbuf) (lexeme_end_p lexbuf)

}

let lc = ['a'-'z']
let uc = ['A'-'Z']
let digit = ['0'-'9']
let any = digit | lc | uc | '_' | '\''
let filename_allowed = any | '.' | '\\' | '/' | '-'

let lcase = lc any*
let ucase = uc any*
let number = digit+
let filename = filename_allowed+

rule token = parse
  [' ' '\t']               { token lexbuf }
| '\n'                     { inc_lines lexbuf; token lexbuf }
| '?' (['a'-'z']+ as x)    { update_lexeme lexbuf; RAW_EVAR(x) }
| lcase as x               { try
                               update_lexeme lexbuf;
                               Hashtbl.find keyword_table x
			     with
			       Not_found -> LCASE(x) }
| ucase as x               { update_lexeme lexbuf; UCASE(x) }
| '%' (lcase as x)         { update_lexeme lexbuf; DIRECTIVE(x) }
| '"' (filename as x) '"'  { update_lexeme lexbuf; QUOTE_STRING(x) }
| number as x              { update_lexeme lexbuf; INTLIT(int_of_string x)}
| "="                      { update_lexeme lexbuf; EQUALS }
| "|"                      { update_lexeme lexbuf; ALT }
| ":"                      { update_lexeme lexbuf; COLON }
| ";"                      { update_lexeme lexbuf; SEMICOLON }
| ";;"                     { update_lexeme lexbuf; DUALSEMICOLON }
| "("                      { update_lexeme lexbuf; LPAREN }
| ")"                      { update_lexeme lexbuf; RPAREN }
| ","                      { update_lexeme lexbuf; COMMA }
| "."                      { update_lexeme lexbuf; DOT }
| "->"		           { update_lexeme lexbuf; ARROW }
| "*"                      { update_lexeme lexbuf; CROSS }
| "<"                      { update_lexeme lexbuf; LANGLE }
| ">"                      { update_lexeme lexbuf; RANGLE }
| "["                      { update_lexeme lexbuf; LBRAC }
| "]"                      { update_lexeme lexbuf; RBRAC }
| "||"                     { update_lexeme lexbuf; BRANCH }
| "&"                      { update_lexeme lexbuf; AMP }
| "#"                      { update_lexeme lexbuf; FRESHFOR }
| "{{"                     { update_lexeme lexbuf; BEGINDEF }
| "}}"                     { update_lexeme lexbuf; ENDDEF }
| "<:"                     { update_lexeme lexbuf; SUBSET }
| "_"                      { update_lexeme lexbuf; UNDERSCORE }
| "[|"                     { update_lexeme lexbuf; BEGINTREE }
| "|]"                     { update_lexeme lexbuf; ENDTREE }
| ['-']+                   { update_lexeme lexbuf; RULELINE }
| "(*"                     { let span = get_span lexbuf in comments [span] lexbuf }
| eof                      { EOF }
| _                        { let span = get_span lexbuf in 
                             raise (IllegalToken span) }

and comments openings = parse
  "*)"                     { match openings with
                               [] -> let span = get_span lexbuf in
                                     raise (IllegalToken span)
                            | [o] -> token lexbuf
                            | o::os -> comments os lexbuf }
| "(*"                    { let span = get_span lexbuf in
                            comments (span::openings) lexbuf }
| eof                     { let span = get_span lexbuf in
                            raise (CommentNotTerminated span) }
| '\n'		          { inc_lines lexbuf; comments openings lexbuf }
| _                       { comments openings lexbuf }
