(* A collection of useful little functions. *)

open Globals

(* If in verbose mode, print out the given string on its own line. If not, then don't! *)
let debug x =
  if !Globals.verbose then
    ((if (!Globals.logfilemode) then
       if (!Globals.logfilechan <> stdout) then
      output_string (!Globals.logfilechan) (x ^ "\n"));
     print_endline x)

let debug_nl () = debug ""


(* Print output to standard output *)
let output x =
  (if (!Globals.logfilemode) then
    if (!Globals.logfilechan <> stdout) then
      output_string (!Globals.logfilechan) (x ^ "\n"));
  print_endline x

let log_output x =
  if (!Globals.logfilemode) then
    if (!Globals.logfilechan <> stdout) then
      output_string (!Globals.logfilechan) x

let output_nl () = output ""


(* Return the first n elements of a list,
 * or the whole list if n > (List.length xs). *)
let rec firstN lst n =
  match (lst,n) with
  _,0 -> []
| [],n -> []
| (x::xs),n -> x::(firstN xs (n-1))

(* Split the list into first n elements and rest. *)
let rec splitN lst n =
  match (lst,n) with
  xs,0 -> [],xs
| [],n -> failwith "splitN - not enough elements in list"
| (x::xs),n ->
    let nxt,rst = splitN xs (n-1) in
    (x::nxt),rst

(* Find the last element of a list *)
let final lst = List.hd (List.rev lst)

(* Project out of a triple *)
let fst3 = fun(x,_,_) -> x
let mid3 = fun(_,x,_) -> x
let last3 = fun(_,_,x) -> x

(* Link and Map *)
let linkMap f xs = List.flatten (List.map f xs)

(* Get out of an option type... *)
let unoption = function
  Some x -> x
| None -> failwith "unoption"

(* string_of_list *)
let string_of_list f sep xs =
  let rec inner xs =
    match xs with
      [] -> ""
    | [x] -> f x
    | x::xs -> ((inner xs) ^ sep ^ (f x))
  in
  inner (List.rev xs)

(* string_of_enclist.
   This allows 2 optional arguments for initial and final delimiters.
   If the list is empty, the delimiters are printed (e.g. for the empty set). *)
let string_of_enclist init final f sep xs = init ^ (string_of_list f sep xs) ^ final

(* string_of_enclist2.
   This allows 2 optional arguments for initial and final delimiters.
   If the list is empty, the empty string is returned. *)
let string_of_enclist2 init final f sep xs =
  match xs with [] -> ""
    | _ -> init ^ (string_of_list f sep xs) ^ final

(* Find the index of a given element in a list (first element of list is 0). *)
let index x l =
  let rec scan x l acc = match l with
    [] -> None
  | (y::ys) -> if (x=y) then Some acc else scan x ys (acc+1)
  in
  scan x l 0

(* Find the second element of the first pair in the list whose first element
    matches the argument (first element of list is 0). *)
let get2 x l =
  let rec scan x l acc = match l with
    [] -> None
  | (y::ys) -> if (x=fst y) then Some (snd y) else scan x ys (acc+1)
  in
  scan x l 0

(* Return the first duplicate element encountered in a list, or None *)
let findDup lst =
  let rec inner lst checked =
    match lst with
      [] -> None
    | (x::xs) -> if (List.mem x xs) || (List.mem x checked) then
        Some x else inner xs (x::checked)
  in
  inner lst []

(* Skip *)
let skip = function () -> ()

(* Identity *)
let id = function x -> x

(* Raise one integer to the power of another *)
let rec power a = function
  0 -> 1
| n -> a*(power a (n-1))

(* Convert a char to a string *)
let string_of_char c = String.make 1 c

(* Convert a string to a char list *)
let charlist_of_string s =
  (let acc = ref [] in
   String.iter (fun c -> acc := (!acc)@[c]) s;
   !acc)

(* Convert a char list to a string *)
let string_of_charlist cs =
  let slist = List.map string_of_char cs in
  String.concat "" slist

(* An "xyzstring" is a string of the form 'a-z'+.
 * These are used to represent internally-generated names.
 * For simplicity, the interpreter represents them internally using integers
 * (starting at 0), and the following functions convert them to strings
 * to make life a bit easier for the user. *)

(* Convert an xyzstring to an int *)
let int_of_xyzstring (x:string) : int = 
  let revchars = List.rev (charlist_of_string x) in
  if not(List.for_all (fun c -> 'a' <= c && c <= 'z') revchars) then
    failwith "ERROR - evarstring not all a-z!"
  else
    let (_,res) = List.fold_left (fun (i,tot) c ->
      let tot = tot + ((power 26 i)*(Char.code c - 96))  in
      (i+1,tot)) (0,0) revchars
    in (res - 1)

(* Convert an int to an xyzstring *)
let xyzstring_of_int (i:int) : string =
  let makeStr j = let n = (((j-1) mod 26)+97) in string_of_char (Char.chr n) in
  let rec do_step j acc =
    match j/26 with
      0 -> let x = makeStr j in (x ^ acc)
    | n -> (let x = makeStr j in
            do_step n (x ^ acc))
  in do_step (i+1) ""

(* Remove duplicates *)
let remDup s =
  let rec do_scan s acc = match s with
    [] -> acc
  | (a::b) -> if List.mem a acc then do_scan b acc
              else do_scan b (a::acc)
  in
  List.rev(do_scan s [])

(* Cons an element onto a list, only if its not there already! *)
let maybeCons x xs = if (List.mem x xs) then xs else x::xs

(* Merge two lists, removing duplicates. *)
let merge xs ys = remDup (xs@ys)

(* Remove duplicates - with a supplied comparison function,
   which should be "true" if the same or "false" otherwise. *)
let remDup2 cmp s =
  let rec do_scan s acc = match s with
    [] -> acc
  | (a::b) -> if (List.exists (cmp a) acc) then do_scan b acc
      else do_scan b (a::acc)
  in
  List.rev(do_scan s [])

(* Merge two lists, removing duplicates.
 * Use the supplied comparison function to check for duplicates. *)
let merge2 cmp xs ys = remDup2 cmp (xs@ys)
