(** Datatypes for location data, used during parsing to (hopefully) give accurate error reporting. *)

open Lexing
open Library

(** Type identifying a location *)
type t = { line : int; character : int }

(** Type identifying a span across 2 locations *)
type span = { start : t; finish : t }

(** A dummy location and span, useful for defining source-to-source translations. *)
let dummy_loc = { line = 0; character = 0}
let dummy_span = { start = dummy_loc; finish = dummy_loc }

(** Create a location from a lexing position *)
let loc_from_pos pos = { line = pos.pos_lnum; character = pos.pos_cnum - pos.pos_bol }

(** Create a span from 2 lexing positions *)
let span_from_poss pos1 pos2 = { start = (loc_from_pos pos1); finish = (loc_from_pos pos2) }

(** Create a span from 2 locations *)
let span_from_locs l1 l2 = { start = l1; finish = l2 }

(** Merge two spans *)
let mergeSpans sp1 sp2 = { start = sp1.start; finish = sp2.finish }

(** Highlight where a span would appear in the console, if it's a single line *)
let highlight sp = if (sp.start.line = sp.finish.line) then
  output ((String.make (sp.start.character + 2) ' ') ^
    (String.make (sp.finish.character - sp.start.character) '^'))
  else ()

(** to_string functions *)
let to_string loc =
  "line " ^ (string_of_int loc.line) ^ ", character " ^ (string_of_int (loc.character + 1))
let span_to_string sp =
  if ((sp.start.line) = (sp.finish.line)) then
    "line " ^ (string_of_int sp.start.line) ^ ", characters " ^
    (string_of_int (sp.start.character + 1)) ^ "-" ^ (string_of_int (sp.finish.character + 1))
  else
    (to_string sp.start) ^ " to " ^ (to_string sp.finish)

(** Imperative state storing the locations of the last lexeme seen.
  * Should help to give meaningful syntax error messages, as ocamlyacc is less than helpful in that respect! *)
let curr_lexeme_start = ref { line = 1; character = 0 }
let curr_lexeme_finish = ref { line = 1; character = 0 }
let curr_lexeme_reset () =
  curr_lexeme_start := { line = 1; character = 0 };
  curr_lexeme_finish := { line = 1; character = 0 }
