open Library

(* Types for abstract machine instructions etc.
 * Based on Landin's SECD machine [http://en.wikipedia.org/wiki/SECD_machine]
 * and inspired by Andrej Bauer's miniml code [http://andrej.com/plzoo]
 *)

type mvalue =
  MVar of int
| MEvar of int
| MUnit
| MYes
| MData of string * mvalue
| MTuple of mvalue list
| MAbs of mvalue * mvalue
| MClosure of frame * environment

and instr =
  IVar of int
| IEvar of int
| IUnit
| IYes
| IData of string
| ITuple of int
| IAbs
| IClosure of frame
| IRecClosure of frame
| IBranch of frame list
| ICall
| IPopEnv
| IDiscardStkTop
| IDiscardEnvTop
| IStkToEnv
| IExists of Types.ty
| IProj of int
| ICase of (string * frame) list
| IFail
| IEcon
| IFcon
| IFresh of string
| IDistinct of int
| IPrint of string

(* A frame is a list of instructions. *)
and frame = instr list

(* An environment is a list of machine values. *)
and environment = mvalue list

(* The stack is another list of machine values. *)
and stack = mvalue list

(* Display the string representation of the evar number i. *)
let string_of_evar i = "?" ^ (xyzstring_of_int i)

(* General version of soMValue, which allows a formatting function for evars. *)
let rec soMValueGeneral f = function
  MVar i -> failwith "SHOULDN'T TRY TO PRETTY-PRINT MACHINE VARIABLES!"
| MEvar i -> f i
| MUnit -> "()"
| MYes -> "yes"
| MData (k,mv) -> k ^ " " ^ (soMValueGeneral f mv)
| MTuple mvs -> "(" ^ (string_of_list (soMValueGeneral f) ", " mvs) ^ ")"
| MAbs (mv,mv') -> "<" ^ (soMValueGeneral f mv) ^ ">" ^ (soMValueGeneral f mv')
| MClosure (_,_) -> "<fun>"

(* Produce a string from a machine value, for debugging purposes. *)
let soMValue mv = soMValueGeneral string_of_evar mv

(* Predicate whether a particular evar appears in a constraint! *)
(*
let containsX i = function
  EQ((_,_,mx),(_,_,my)) -> ...
| FR(_,(_,_,my)) -> ...
*)

(* Find out whether a machine value is an evar or not. *)
let isEvar = function MEvar _ -> true | _ -> false

(* Get the list of all integers which appear as evars in a machine value! *)
let rec evars = function
  MEvar i -> [i]
| MVar _ | MUnit | MYes | MClosure _ -> []
| MData (_,mv) -> evars mv
| MTuple mvs -> List.fold_left (fun xs mv -> merge xs (evars mv)) [] mvs
| MAbs (mv,mv') -> merge (evars mv) (evars mv')

(* Produce a string representation of a machine instruction. *)
let rec soInstr = function
  IVar n -> "IVar(" ^ (string_of_int n) ^ ")"
| IEvar n -> "IEVar(" ^ (string_of_int n) ^ ")"
| IUnit -> "IUnit"
| IYes -> "IYes"
| IData k -> "IData(" ^ k ^ ")"
| ITuple n -> "ITuple(" ^ (string_of_int n) ^ ")"
| IAbs -> "IAbs"
| IClosure frm ->
    "IClosure(" ^ (string_of_enclist "[" "]" soInstr ";" frm) ^ ")"
| IRecClosure frm ->
    "IRecClosure(" ^ (string_of_enclist "[" "]" soInstr ";" frm) ^ ")"
| IBranch frms -> "IBranch(" ^ (string_of_list soFrame "," frms) ^ ")"
| ICall -> "ICall"
| IPopEnv -> "IPopEnv"
| IDiscardStkTop -> "IDiscardStkTop"
| IDiscardEnvTop -> "IDiscardEnvTop"
| IStkToEnv -> "IStkToEnv"
| IExists t -> "IExists(" ^ (Types.soType t) ^ ")"
| IProj n -> "IProj(" ^ (string_of_int n) ^ ")"
| ICase kfrms ->
    let soClause (k,frm) = "{" ^ k ^ "," ^ (soFrame frm) ^ "}" in
    "ICase(" ^ (string_of_list soClause "," kfrms) ^ ")"
| IFail -> "IFail"
| IEcon -> "IEcon"
| IFcon -> "IFcon"
| IFresh n -> "IFresh(" ^ n ^ ")"
| IDistinct n -> "IDistinct(" ^ (string_of_int n) ^ ")"
| IPrint x -> "IPrint(" ^ x ^ ")"
and soFrame frm = string_of_enclist "[" "]" soInstr ";" frm
let soEnv env = string_of_enclist "[<" ">]" soMValue ";" env
let soEnvs envs = string_of_list soEnv "\n" envs
