open Library
open Graph
open Errors
open Types
open Machine
open Constraint

(* ************************************************************ *)
(* Graph implementation for representing name-name constraints. *)
(* ************************************************************ *)

(* An imperative graph datatype for recording equalities and freshnesses
 * between variables of NAME SORT only. *)
module G = Imperative.Graph.Abstract(struct type t = (int list) ref end)

(* Exceptions raised if try to:
 * 1. add an edge that already exists,
 * 2. add an edge for a vertex/vertices that don't exist,
 * 3. remove a vertex that still has edges,
 * 4. remove a vertex that doesn't exist,
 * 5. remove an edge that doesn't exist,
 * 6. add a vertex that already exists. *)
exception EdgeAlreadyExists of G.V.t * G.V.t
exception EdgeVertexNotFound of G.V.t
exception VertexStillConnected of G.V.t
exception VertexNotFound of G.V.t
exception EdgeNotFound of G.V.t * G.V.t
exception VertexAlreadyExists of G.V.t

(* The reference used to store the current graph. *)
let g = ref (G.create ())

(* Produce an evar-like string, given an integer *)
let soi = string_of_evar

(* Produce a string representation of a vertex *)
let string_of_vertex v =
  let vars = !(G.V.label v) in
  "{" ^ (string_of_list soi "," vars) ^ "}"

(* Print the current name-name constraint graph. *)
let printNameGraph () =
 (let printVertex v =
   debug (string_of_vertex v)
  in
  let printEdge v1 v2 =
    debug ((string_of_vertex v1) ^ "<--->" ^
                   (string_of_vertex v2))
  in
  debug "==============";
  debug "Vertices are:";
  G.iter_vertex printVertex !g;
  debug "Edges are:";
  G.iter_edges printEdge !g;
  debug "==============")

(* A type for the different kinds of operations that we might
 * want to apply to g.
 * This will allow us to compute the "inverse" of a particular
 * operation, for the purposes of easily rolling back changes
 * made to the graph, avoiding lots of unnecessary deep copying
  * of the graph. *)
type graphOp = AddVertex of G.V.t
             | AddEdge of G.V.t * G.V.t
             | RelabelVertex of G.V.t * (int list)
             | RemoveVertex of G.V.t
             | RemoveEdge of G.V.t * G.V.t

(* Produce a string representation of a graphOp *)
let string_of_graphOp = function
  AddVertex v -> ("AddVertex " ^ (string_of_vertex v))
| AddEdge (v1,v2) -> ("AddEdge " ^ (string_of_vertex v1) ^
    "<--->" ^ (string_of_vertex v2))
| RelabelVertex (v,xs) ->
    ("RelabelVertex " ^ (string_of_vertex v) ^ " => {" ^
    (string_of_list soi "," xs) ^ "}")
| RemoveVertex v -> ("RemoveVertex " ^ (string_of_vertex v))
| RemoveEdge (v1,v2) -> ("RemoveEdge " ^ (string_of_vertex v1) ^
    "<--->" ^ (string_of_vertex v2))

(* Elements of the "undo" stack will either be graph operations
 * or "marks", which tell us when to stop unwinding the stack
 * and provide a thunk to continue searching. *)
type opMark = GOP of graphOp
            | MARK

(* Produce a string representation of an opMark *)
let string_of_opMark = function
  GOP gop -> "GOP: " ^ (string_of_graphOp gop)
| MARK -> "MARK"

(* A list of "undo" operations which, if popped and performed
 * in order, will roll back the mutations performed on the graph
 * and return it to its original state.
 * As with graphs, we only have one undo list active. *)
let undoList : (opMark list) ref = ref []

(* Print the current contents of the undo list. *)
let printUndoList () =
  let ctr = ref 1 in
  List.iter (fun opm -> let i = !ctr in ctr := i+1;
    debug ((string_of_int i) ^ ". " ^ (string_of_opMark opm)))
    !undoList

(* Add an undo operation to the undo list. *)
let queueUndo op = undoList := ((GOP op) :: !undoList)

(* Add a mark to the undo list, with thunk f. *)
let queueMark () = undoList := (MARK :: !undoList)

(* Reset the graph to its initial state. *)
let resetNDState () = (g := (G.create ()); undoList := [])

(* Apply a graph operation to the graph, compute the "inverse"
 * operation and queue it in the "undo" stack. *)

let apply gop =
 (debug ("EXECUTING: " ^ (string_of_graphOp gop));
  match gop with
  (* Add a new vertex. *)
  AddVertex v ->
    (* Check that the edge doesn't already exist and raise an exception if
     * necessary, so our operations don't have unexpected side effects! *)
    if G.mem_vertex !g v then
      raise (VertexAlreadyExists v)
    else
     (queueUndo (RemoveVertex v);
      G.add_vertex !g v)
  (* Add a new edge between two existing vertices. *)
| AddEdge (v1,v2) ->
    (* Check that the edge doesn't already exist and that both vertices
     * are already in the graph and raise exceptions if
     * necessary, so our operations don't have unexpected side effects! *)
    if (G.mem_edge !g v1 v2) then
      raise (EdgeAlreadyExists (v1,v2))
    else
      (match (G.mem_vertex !g v1),(G.mem_vertex !g v2) with
         true,true ->
          (queueUndo (RemoveEdge(v1,v2));
           G.add_edge !g v1 v2)
       | true,false ->
           raise (EdgeVertexNotFound v2)
       | false,_ ->
           raise (EdgeVertexNotFound v1))
  (* Change the label on a vertex. *)
| RelabelVertex (v,xs) ->
   (let vref = G.V.label v in
    let oldLabel = !vref in
    queueUndo (RelabelVertex(v,oldLabel));
    vref := xs)
  (* Remove a vertex from the graph. *)
| RemoveVertex v ->
    (* Check that the vertex is already in the graph and isn't
     * still connected and raise exceptions if necessary,
     * so our operations don't have unexpected side effects! *)
    if (G.mem_vertex !g v) then
      (match (G.succ !g v) with
         [] -> (queueUndo(AddVertex v);
                G.remove_vertex !g v)
       | _ -> raise (VertexStillConnected v))
    else
      raise (VertexNotFound v)
  (* Remove an edge from the graph. *)
| RemoveEdge (v1,v2) ->
    (* Check that the edge exists and raise an exception if necessary,
     * so our operations don't have unexpected side effects! *)
    if (G.mem_edge !g v1 v2) then
     (queueUndo (AddEdge(v1,v2));
      G.remove_edge !g v1 v2)
    else
      raise (EdgeNotFound (v1,v2)))

(* Apply a list of graph operations, starting from the head. *)
let applyList = List.iter apply

(* Unwind changes made to the graph, up to the last choice point.
 * This basically involves popping elements of the "undo" list and applying
 * them, until we reach a MARK at which point we stop
 * unwinding and execute the continuation contained in the MARK. *)
let unwind () =
  (* Apply a graph operation to the graph, without bothering
   * about recording undo information. *)
  let applyNoUndo gop =
   (debug ("Unwinding: " ^ (string_of_graphOp gop));
    match gop with
    AddVertex v -> G.add_vertex !g v
  | AddEdge (v1,v2) -> G.add_edge !g v1 v2
  | RelabelVertex (v,xs) -> let vref = G.V.label v in vref := xs
  | RemoveVertex v -> G.remove_vertex !g v
  | RemoveEdge (v1,v2) -> G.remove_edge !g v1 v2)
  in
  (* Actually unwind the graph... *)
  let rec doRewind = function
    [] -> failwith ("*** SHOULDN'T BE ABLE TO REWIND PAST " ^ 
                    "BEGINNING OF UNDO LIST ***")
    (* Apply an undo operation, but don't record any
     * undo information for the undo operation! *)
  | (GOP op)::rest -> (applyNoUndo op; doRewind rest)
    (* When find a mark, remove it and save the new undo list. *)
  | MARK::rest -> undoList := rest
  in
  doRewind !undoList

(* Exception used to find variables in graph. *)
exception FoundVariableIn of G.V.t

(* Find the vertex which has x in its label, in the given graph.
 * The function actually returns Some v if v is the
 * first such vertex that it finds (which should be the only
 * such vertex), or None if so no such vertex exists. *)
let findVarGeneral grph x =
  try
    (G.iter_vertex (fun v -> let vars = G.V.label v in
                    if (List.mem x (!vars)) then
                    raise (FoundVariableIn v)) grph; None)
  with
    FoundVariableIn v -> Some v

(* Find the vertex which has x in its label, in the global graph g. *)
let findVar x = findVarGeneral !g x

(* Return a new vertex with given list of integers *)
let createVertex xs = G.V.create (ref xs)

(* *********************************************************** *)
(*   Code for actually processing name-name constraints.       *)
(* *********************************************************** *)

(* Add a freshness between two variables of name sort *)
let addNameFreshness x y =
  if x=y then
    (debug ("FRESHNESS FAILURE when processing " ^
      (soi x) ^ " # " ^ (soi y));
    raise ConstraintBacktrack)
  else match (findVar x),(findVar y) with
    None,None ->
      (* Create two new singleton vertices, add them to the graph
       * and constrain them to be distinct. *)
     (let vx,vy = (createVertex [x]),(createVertex [y]) in
      applyList [AddVertex vx; AddVertex vy; AddEdge (vx,vy)])
  | (Some vx), None ->
      (* Add a new singleton vertex and constrain it to be
       * distinct from the existing vertex. *)
     (let vy = (createVertex [y]) in
      applyList [AddVertex vy; AddEdge (vx,vy)])
  | None, (Some vy) ->
      (* Add a new singleton vertex and constrain it to be
       * distinct from the existing vertex. *)
     (let vx = (createVertex [x]) in
      applyList [AddVertex vx; AddEdge (vx,vy)])
  | (Some vx),(Some vy) ->
       (* If the same vertex then fail. *)
       if (G.V.equal vx vy) then
         (debug ("FRESHNESS FAILURE when processing " ^
            (soi x) ^ " # " ^ (soi y));
          raise ConstraintBacktrack)
       else
        (if (G.mem_edge !g vx vy) then
           (* Vertices already constrained to be distinct - do nothing. *)
           ()
         else
           (* Add an edge between the two vertices. *)
           apply(AddEdge (vx,vy)))

(* Add an equality between two variables of name sort *)
let addNameEquality x y =
  match (findVar x),(findVar y) with
    None,None ->
      (* Add a new vertex with the two variables in.*)
     (let vxy = createVertex [x;y] in
      apply(AddVertex vxy))
  | (Some vx), None ->
      (* Add y to the existing vertex. *)
     (let vars = G.V.label vx in
      let newList = remDup ((!vars) @ [y]) in
      apply(RelabelVertex (vx,newList)))
  | None, (Some vy) ->
      (* Add x to the existing vertex. *)
     (let vars = G.V.label vy in
      let newList = remDup ((!vars) @ [x]) in
      apply(RelabelVertex (vy,newList)))
  | (Some vx),(Some vy) ->
      if (G.V.equal vx vy) then
        (* Do nothing if already in the same vertex. *)
        ()
      else
        if (G.mem_edge !g vx vy) then
          (* If the two vertices are constrained to be
           * distinct, then fail. *)
          (debug ("EQUALITY FAILURE when processing " ^
             (soi x) ^ " = " ^ (soi y));
           raise ConstraintBacktrack)
        else
          (* Combine the two lists of variables and create
           * a new vertex with this list. *)
         (let xvars = G.V.label vx in
          let yvars = G.V.label vy in
          let combinedList = remDup ((!xvars) @ (!yvars)) in
          let vxy = createVertex combinedList in
          (* Apply the following operations to the graph:
           * 1. Add the new vertex vxy to the graph.
           * 2. Remove edges to the old vertices (vx,vy) and add edges
           *    to the new vertex vxy instead.
           * 3. Remove the old vertices (vx,vy). *)
          let vxSuccs = G.succ !g vx in
          let vySuccs = G.succ !g vy in
          let vxySuccs = remDup (vxSuccs @ vySuccs) in
          let vxRemOps = List.map (fun v -> RemoveEdge(v,vx)) vxSuccs in
          let vyRemOps = List.map (fun v -> RemoveEdge(v,vy)) vySuccs in
          let remOps = vxRemOps @ vyRemOps in
          let addOps = List.map (fun v -> AddEdge(v,vxy)) vxySuccs in
          applyList ([AddVertex vxy] @ remOps @ addOps @
                     [RemoveVertex vx; RemoveVertex vy]))

(* A type for "solutions" which are conversions of the imperative graph into
 * a format that can be used to output meaningful results to the user. *)
type solution = G.t (* *** OKAY??? *** *)

(* Function to convert the current imperative graph into a "solution". *)
let convertGraph () = G.copy (!g) (* *** OKAY??? *** *)

(* Given a solution, get the "freshfor" variables of some int. *)
let getFreshFors sln i =
  match (findVarGeneral sln i) with
    None -> []
  | Some vi ->
      let succs = G.succ sln vi in
      remDup(linkMap (fun vtx -> !(G.V.label vtx)) succs)

(* Given a solution, get the "equal to" variables of some int. *)
let getEquals sln i =
  match (findVarGeneral sln i) with
    None -> []
  | Some vi ->
      let varlist = !(G.V.label vi) in
      List.filter (fun x -> x <> i) varlist

(* Explore the satisfiability of name-name constraints using non-deterministic
 * search. The imperative graph data structure is used to detect inconsistencies. *)
let nameConstraintSearch (vd:Vardata.t) (cs:prob) (enum:bool) =
  begin
    (* Reset the graph and "undo list" to be empty initially! *)
    resetNDState ();

    (* Initialise a reference for storing the list of solutions that we compute. *)
    let solns = ref [] in

    (* Quick function for looking up the type of a variable. *)
    let typeOf i = Vardata.getType vd i in

    (* Process the constraints with the fewest abstractions first.
     * These are more constrained and should therefore reduce the number
     * of branches to explore when processing the longer constraints. *)
    let sortingFn c1 c2 =
      match(c1,c2) with
        (EQ((n1,_,_),_),EQ((n2,_,_),_))
      | (EQ((n1,_,_),_),FR(_,(n2,_,_)))
      | (FR(_,(n1,_,_)),EQ((n2,_,_),_))
      | (FR(_,(n1,_,_)),FR(_,(n2,_,_))) ->
          compare n1 n2
    in
    let cs = List.sort sortingFn cs in

    (* Do the branching search, adding edges to the graph as appropriate.
     * Use MARKS with thunks to redirect control flow when backtrack. *)
    let rec doNDSearch = function

      (* ********* *)
      (* FINISHED? *)
      (* ********* *)
      [] ->
	if enum then
	  begin
	    solns := (convertGraph ())::(!solns);
	    raise ConstraintBacktrack
	  end
	else raise ConstraintSuccess

      (* ********************* *)
      (* EQUALITY CONSTRAINTS! *)
      (* ********************* *)
    | ((EQ((nx,xs,(MEvar x)),(ny,ys,(MEvar y))))::cs) ->
        (let debug_init () =
           begin
             debug "\n**************************\n";
             debug ("PROCESSING: " ^
               (string_of_constr vd (EQ((nx,xs,(MEvar x)),(ny,ys,(MEvar y))))))
           end
         in
         match (typeOf x),(typeOf y) with
           (Nty n),(Nty n') ->
              if (n=n') then
		begin
		  if (x=y) then
		    (* Rule E6 *)
                    doNDSearch cs
		  else
		    (* Rule E5 *)
		    let rec eqBranch = function
			0 -> begin
  			  debug_init ();
			  debug "Entered eqBranch at depth 0";
			  printNameGraph ();
			  (* Add freshnesses for any names in the lists xs,ys. *)
			  List.iter (fun z ->
                            addNameFreshness x z;
                            debug ((soi x) ^ " should be freshfor " ^
                              (soi z))) xs;
			  List.iter (fun z ->
                            addNameFreshness y z;
                            debug ((soi y) ^ " should be freshfor " ^
                              (soi z))) ys;
			  (* x should be equal to y. *)
			  addNameEquality x y;
			  (* Process the rest of the constraints. *)
			  doNDSearch cs
			end
                      | n -> begin
			  debug_init ();
			  debug ("Entered eqBranch at depth " ^
                            (string_of_int n));
			  printNameGraph ();
			  let i = nx - n in
			  (* What should x and y be fresh for / equal? *)
			  let xFreshFor = firstN xs i in
			  let yFreshFor = firstN ys i in
			  let xEq = List.nth xs i in
			  let yEq = List.nth ys i in
			  (* Check that the types match when we impose equalities. *)
			  if (typeOf x = typeOf xEq) then
                            (try
                              (* Put a mark in the undo list. *)
                              (queueMark ();
                              (* Add the appropriate freshnesses. *)
                              List.iter (fun z ->
				addNameFreshness x z;
				debug ((soi x) ^ " should be freshfor " ^
				  (soi z))) xFreshFor;
                              List.iter (fun z ->
				addNameFreshness y z;
				debug ((soi y) ^ " should be freshfor " ^
				  (soi z))) yFreshFor;
                              (* Add the appropriate equalities. *)
                              addNameEquality x xEq;
                              debug ((soi x) ^ " should equal " ^ (soi xEq));
                              addNameEquality y yEq;
                              debug ((soi y) ^ " should equal " ^ (soi yEq));
                              (* Process the rest of the constraints. *)
                              doNDSearch cs)
                            with ConstraintBacktrack ->
                              begin
				(* Unwind the graph to our mark. *)
				unwind ();
				(* Resume searching with a recursive call. *)
				eqBranch (n-1)
                              end)
                      else
                        (* Since the types are different, these two can't be
                         * made equal, so try the next abstraction in the list. *)
                        eqBranch (n-1)
                    end
		  in
		  eqBranch nx
		end
            else failTC "Equality constraint solver - 1"
       | _,_ -> failTC "Equality constraint solver - 2")

      (* ********************** *)
      (* FRESHNESS CONSTRAINTS! *)
      (* ********************** *)
    | ((FR(x,(ny,ys,(MEvar y))))::cs) ->
      (let debug_init () =
         begin
           debug "\n**************************\n";
           debug ("PROCESSING: " ^
             (string_of_constr vd (FR(x,(ny,ys,(MEvar y))))))
         end
      in
      match (typeOf x),(typeOf y) with
        (Nty ntyx),(Nty ntyy) ->
           let rec frBranch = function
               0 -> begin
		 if (ntyx = ntyy) then
		   begin
		     (* Add a freshness to the graph. *)
                     debug_init ();
                     debug "Entered frBranch at depth 0";
                     printNameGraph ();
                     (* Add the appropriate freshnesses. *)
                     List.iter (fun z ->
                       addNameFreshness x z;
                       debug ((soi x) ^ " should be freshfor " ^
			 (soi z))) ys;
                     (* x should be fresh for y. *)
                     addNameFreshness x y;
                     (* Process the rest of the constraints. *)
                     doNDSearch cs
		   end
		 else
		   (* Rule F6 *)
		   doNDSearch cs
               end
           | n -> begin
	       (* Rule F6 *)
                    debug_init ();
                    debug ("Entered frBranch at depth " ^
                      (string_of_int n));
                    printNameGraph ();
                    let i = ny - n in
                    (* What should x be fresh for / equal? *)
                    let xFreshFor = firstN ys i in
                    let xEq = List.nth ys i in
                    (* Check that the types match when we impose equalities. *)
                    if (typeOf x = typeOf xEq) then
                      (try
                        (* Put a mark in the undo list. *)
                        (queueMark ();
                        (* Add the appropriate freshnesses. *)
                        List.iter (fun y ->
                          addNameFreshness x y;
                          debug ((soi x) ^ " should be freshfor " ^
                            (soi y))) xFreshFor;
                        (* Add the appropriate equality. *)
                        addNameEquality x xEq;
                        debug ((soi x) ^ " should equal " ^ (soi xEq));
                        (* Process the rest of the constraints. *)
                        doNDSearch cs)
                      with ConstraintBacktrack ->
                        (* Unwind the graph to our mark. *)
                        (unwind ();
                        (* Resume searching with a recursive call. *)
                        frBranch (n-1)))
                    else
                      (* Since the types are different, these two can't be
                       * made equal, so try the next abstraction in the list. *)
                      frBranch (n-1)
                  end
           in
           frBranch ny
       | _ -> failTC "Freshness constraint solver")

      (* ******************** *)
      (* ANYTHING ELSE FAILS! *)
      (* ******************** *)
    | _ -> failTC "Non-evar constraint in namegraph!"
    in
    try doNDSearch cs with
      ConstraintBacktrack -> if enum then (List.rev !solns) else raise ConstraintBacktrack
  end

(* Check satisfiability of name-name constraints. *)
let checkSat (vd:Vardata.t) (cs:prob) =
  try ignore(nameConstraintSearch vd cs false); failwith "Error - checkSat" with
      ConstraintBacktrack -> begin debug "Namegraph failure."; false end
    | ConstraintSuccess   -> begin debug "Namegraph success."; true  end

(* Return a list of "solutions" to some name-name constraints. *)
let enumSolns (vd:Vardata.t) (cs:prob) =
   nameConstraintSearch vd cs true
