/* Parser for the alphaML language */

%{
 open Ast
 open Cmd
 open Library
 open Location
 open Lexing
 open Parsing
 open Errors

 (* Get the current "location" from a lexbuf - for throwing syntax error exceptions. *)
 let get_pos lexbuf =
   let l = lexbuf.lex_curr_p.pos_lnum in
   let c = lexbuf.lex_curr_p.pos_cnum - lexbuf.lex_curr_p.pos_bol in
   { line = l; character = c }

  (* Create a rawId with the given string at the given point in RHS of rule *)
  let createId x n =
    let sp = span_from_poss (rhs_start_pos n) (rhs_end_pos n) in
    if x="rel" then raise (SyntaxError sp) else ((x,sp):rawId)

  (* Create a string, checking it's not "rel"! *)
  let createString x n =
    let sp = span_from_poss (rhs_start_pos n) (rhs_end_pos n) in
    if x="rel" then raise (SyntaxError sp) else x

  (* Create a wildcard identifier. *)
  let createWildcard n =
    let sp = span_from_poss (rhs_start_pos n) (rhs_end_pos n) in
    ((Ast.wildcard,sp):rawId)

  (* Create a rawInt with the given value at the given point in RHS of rule *)
  let createInt i n =
    let sp = span_from_poss (rhs_start_pos n) (rhs_end_pos n) in
    ((i,sp):rawInt)

  (* Create a span across the LHS of a rule *)
  let create_span () =
    span_from_poss (symbol_start_pos ()) (symbol_end_pos ())

  (* Raise a parse error *)
  let parse_error str =
    let sp = span_from_locs (!curr_lexeme_start) (!curr_lexeme_finish) in
    raise (SyntaxError sp)

%}

/* Start points. */

%start main
%start cmd
%start cmdlist
%start searchcmd
%type <Ast.rawExpr> main
%type <Cmd.cmd> cmd
%type <Cmd.cmd list> cmdlist
%type <Cmd.searchCmd> searchcmd

/* Tokens. */

%token ALT
%token AMP
%token AND
%token ARROW
%token AS
%token BEGINDEF
%token BEGINTREE
%token BRANCH
%token CASE
%token COLON
%token COMMA
%token CROSS
%token DATATYPE
%token DISTINCT
%token <string> DIRECTIVE
%token DOT
%token DUALSEMICOLON
%token ENDDEF
%token ENDTREE
%token EOF
%token EQUALS
%token EXISTS
%token FAIL
%token FN
%token FRESH
%token FRESHFOR
%token FUN
%token IN
%token <int> INTLIT
%token LANGLE
%token LBRAC
%token <string> LCASE
%token LET
%token LPAREN
%token NAMETYPE
%token OF
%token OFF
%token ON
%token <string> QUOTE_STRING
%token RANGLE
%token RBRAC
%token RPAREN
%token <string> RAW_EVAR
%token REC
%token REL
%token RELATION
%token RULELINE
%token SEMICOLON
%token SOME
%token SUBSET
%token SUCCEED
%token <string> UCASE
%token UNBIND
%token UNBIND_FRESH
%token UNDERSCORE
%token UNIT
%token WHERE

/* Associativities and precedences. */

%nonassoc simple
%nonassoc let_rule
%nonassoc branches_rule
%nonassoc case_rule
%nonassoc unbind_rule
%left ALT
%nonassoc below_cross
%left CROSS
%right ARROW
%nonassoc abs_rule
%left BRANCH
%left AMP
%nonassoc FRESHFOR
%nonassoc EQUALS
%nonassoc DOT
%left RANGLE
%right UCASE
%nonassoc below_simple
%nonassoc LCASE RAW_EVAR LPAREN SOME FAIL SUCCEED FRESH BEGINDEF UNIT REL LBRAC DISTINCT BEGINTREE

%%
main:
  expr EOF                               { $1 }
;

/* Toplevel interpreter commands. */

searchcmd:
  DOT                                    { Stop }
| COMMA                                  { CarryOn }
;
cmd:
  DIRECTIVE DUALSEMICOLON                { Directive(VoidDir($1)) }
| DIRECTIVE QUOTE_STRING DUALSEMICOLON   { Directive(StringDir($1,$2)) }
| DIRECTIVE ON DUALSEMICOLON             { Directive(BoolDir($1,true)) }
| DIRECTIVE OFF DUALSEMICOLON            { Directive(BoolDir($1,false)) }
| expr DUALSEMICOLON                     { ExprCmd($1) }
| LET lcase_or_wild EQUALS
    expr DUALSEMICOLON                   { LetCmd($2,[],$4) }
| LET LCASE paren_tyasm_list
    EQUALS expr DUALSEMICOLON            { LetCmd((createId $2 2),$3,$5) }
| LET REC LCASE paren_tyasm_list
    COLON typ EQUALS expr DUALSEMICOLON  { LetRecCmd((createId $3 3),$4,$6,$8) }
| DATATYPE muldtydec DUALSEMICOLON       { DtyDecCmd($2) }
| NAMETYPE nsorts DUALSEMICOLON          { NsortDecCmd($2) }
| RELATION mulreldec DUALSEMICOLON       { RelDecCmd($2) }
;
cmdlist:
  cmdlist cmd                            { $1 @ [$2] }
| cmd                                    { [$1] }
;

/* Datatype, nametype relation declarations. */

mulreldec:
  mulreldec AND reldec                   { $1 @ [$3] }
| reldec                                 { [$1] }
;
reldec:
  UCASE SUBSET typ                       { ($1,$3) }
;
muldtydec:
  muldtydec AND dtydec                   { $1 @ [$3] }
| dtydec                                 { [$1] }
;
nsorts:
  nsorts AND LCASE                       { $1 @ [createString $3 3] }
| LCASE                                  { [createString $1 1] }
;
dtydec:
  LCASE EQUALS decrhs                    { ((createString $1 1),$3) }
;
decrhs:
  decrhs ALT singledecrhs                { $1 @ [$3] }
| singledecrhs                           { [$1] }
;
singledecrhs:
  UCASE OF typ                           { ($1,$3) }
;

/* Types. */

typ:
  typ ARROW typ                          { RawFunTy($1,$3) }
| simple_typs %prec below_cross          { match $1 with [] -> failwith "Parser - empty list of types"
                                                       | [t] -> t
                                                       | ts -> RawTupTy(ts) }
;
simple_typs:
  simple_typs CROSS simple_typ           { $1 @ [$3] }
| simple_typ                             { [$1] }
;
simple_typ:
  UNIT                                   { RawUnitTy }
| REL                                    { RawRelTy }
| LCASE                                  { RawTypeId(createString $1 1) }
| LBRAC LCASE RBRAC typ %prec abs_rule   { RawAbsTy((createString $2 2),$4) }
| LPAREN typ RPAREN                      { $2 }
;

/* alphaML metalanguage expressions. */

expr:
  simple_expr %prec below_simple         { $1 }
| simple_expr simple_exprs
    %prec below_simple                   { RawAppExpr(($1::$2),create_span()) }
| LET lcase_or_wild EQUALS
    expr IN expr %prec let_rule          { RawLetExpr($2,[],$4,$6,create_span()) }
| LET LCASE paren_tyasm_list EQUALS
    expr IN expr %prec let_rule          { RawLetExpr((createId $2 2),$3,$5,$7,create_span()) }
| LET REC LCASE paren_tyasm_list
    COLON typ EQUALS expr
    IN expr %prec let_rule               { RawLetRecExpr((createId $3 3),$4,$6,$8,$10,create_span()) }
| CASE simple_expr OF multicase
    %prec case_rule                      { RawCaseExpr($2,$4,create_span()) }
| UCASE expr                             { RawDataExpr((createId $1 1),$2,create_span()) }
| LANGLE expr RANGLE expr                { RawAbsExpr($2,$4,create_span()) }
| FN LPAREN LCASE COLON typ RPAREN
    ARROW expr                           { RawLamExpr((createId $3 3),$5,$8,create_span()) }
| FUN LCASE LPAREN LCASE COLON typ
    RPAREN COLON typ EQUALS expr         { RawRecExpr((createId $2 2),(createId $4 4),$6,$9,$11,create_span()) }
| expr DOT INTLIT                        { RawProjExpr($1,(createInt $3 3),create_span()) }
| expr BRANCH expr_branches              { RawBranchExpr(($1::$3),create_span()) }
| expr AMP expr                          { RawSeqExpr($1,$3,create_span()) }
| expr EQUALS expr                       { RawEconExpr($1,$3,create_span()) }
| expr FRESHFOR expr                     { RawFconExpr($1,$3,create_span()) }
| UNBIND expr AS LANGLE LCASE RANGLE
    LCASE COLON LBRAC LCASE RBRAC
    typ IN expr %prec unbind_rule        { RawUnbindExpr($2,(createId $5 5),(createId $7 7),(createId $10 10),$12,$14,false,create_span()) }
| UNBIND_FRESH expr AS LANGLE LCASE
    RANGLE LCASE COLON LBRAC LCASE
    RBRAC typ IN expr %prec unbind_rule  { RawUnbindExpr($2,(createId $5 5),(createId $7 7),(createId $10 10),$12,$14,true,create_span()) }
| EXISTS LCASE COLON typ DOT expr        { RawExistsExpr((createId $2 2),$4,$6,create_span()) }
;
simple_expr:
  LCASE                                  { RawVarExpr(createId $1 1) }
| RAW_EVAR                               { RawEvarExpr(createId $1 1) }
| LPAREN RPAREN                          { RawUnitExpr(create_span()) }
| SOME typ                               { RawSomeExpr($2,create_span()) }
| FAIL typ                               { RawFailExpr($2,create_span()) }
| SUCCEED                                { RawSuccessExpr(create_span()) }
| FRESH LCASE                            { RawFreshExpr((createId $2 2),create_span()) }
| DISTINCT LPAREN expr_commas RPAREN     { RawDistinctExpr($3,create_span()) }
| BEGINDEF schrules ENDDEF               { RawDefnExpr($2,create_span()) }
| BEGINTREE tree ENDTREE                 { RawTreeExpr($2,create_span()) }
| LPAREN expr_commas RPAREN              { RawTupExpr($2,create_span()) }
| LPAREN expr RPAREN                     { $2 }
;
expr_branches:
  expr_branches BRANCH expr              { $1 @ [$3] }
| expr %prec branches_rule               { [$1] }
;
expr_commas:
  expr_commas COMMA expr                 { $1 @ [$3] }
| expr COMMA expr                        { [$1; $3] }
;
simple_exprs:
  simple_exprs simple_expr               { $1 @ [$2] }
| simple_expr                            { [$1] }
;
case:
  UCASE lcase_or_wild ARROW expr	 { ((createId $1 1),$2,$4) }
;
multicase:
  multicase ALT case                     { $1 @ [$3] }
| case                                   { [$1] }
;
lcase_or_wild:
  LCASE                                  { createId $1 1 }
| UNDERSCORE                             { createWildcard 1 }
;

/* Schematic inductive rules. */

schrules:
  schrules schrule                       { $1 @ [$2] }
| schrule                                { [$1] }
;
schrule:
  expr RULELINE atomic                   { ($1,$3,None,[],create_span()) }
| expr RULELINE
    LBRAC LCASE RBRAC atomic             { ($1,$6,(Some $4),[],create_span()) }
| expr RULELINE
    LBRAC tyasm_commas RBRAC atomic      { ($1,$6,None,$4,create_span()) }
| expr RULELINE LBRAC LCASE
    WHERE tyasm_commas RBRAC atomic      { ($1,$8,(Some $4),$6,create_span()) }
;
tyasm_commas:
  tyasm_commas COMMA multi_tyasm         { $1 @ $3 }
| multi_tyasm                            { $1 }
;
multi_tyasm:
  lcaselist COLON typ                    { List.map (fun x -> (x,$3)) $1 }
;
paren_tyasm_list:
  paren_tyasm_list paren_tyasm           { $1 @ [$2] }
| paren_tyasm                            { [$1] }
;
paren_tyasm:
  LPAREN LCASE COLON typ RPAREN          { ((createId $2 2),$4) }
;
lcaselist:
  lcaselist COMMA LCASE                  { $1 @ [createId $3 3] }
| LCASE                                  { [createId $1 1] }
;
atomic:
  UCASE LPAREN patterns_oneplus RPAREN   { ((createId $1 1),$3,create_span()) }
;
pattern:
  LCASE                                  { RawVarPat(createId $1 1) }
| LPAREN RPAREN                          { RawUnitPat(create_span()) }
| LANGLE LCASE RANGLE pattern            { RawAbsPat((createId $2 2),$4,create_span()) }
| LPAREN pattern_commas RPAREN           { RawTupPat($2,create_span()) }
| UCASE pattern                          { RawDataPat((createId $1 1),$2,create_span()) }
| LPAREN pattern RPAREN                  { $2 }
;
pattern_commas:
  pattern_commas COMMA pattern           { $1 @ [$3] }
| pattern COMMA pattern                  { [$1;$3] }
;
patterns_oneplus:
  pattern_commas_oneplus                 { ($1,create_span()) }
;
pattern_commas_oneplus:
  pattern_commas_oneplus COMMA pattern   { $1 @ [$3] }
| pattern                                { [$1] }
;
tree:
  LCASE                                  { RawNameTree(createId $1 1) }
| LPAREN RPAREN                          { RawUnitTree(create_span()) }
| LANGLE LCASE COLON typ RANGLE tree     { RawAbsTree((createId $2 2),$4,$6,create_span()) }
| LPAREN tree_commas RPAREN              { RawTupTree($2,create_span()) }
| UCASE tree                             { RawDataTree((createId $1 1),$2,create_span()) }
| LPAREN tree RPAREN                     { $2 }
;
tree_commas:
  tree_commas COMMA tree                 { $1 @ [$3] }
| tree COMMA tree                        { [$1;$3] }
;
