(* An implementation of the constraint solver. *)

open Library
open Errors
open Types
open Machine
open Constraint

(* Function to perform the "occurs check" for the constraint solver.
 * Returns true if the occurs check is successful, i.e. the variable
 * DOES NOT occur in the value. *)
let rec occurs_check_succeeds vd x = function
    MEvar y ->
      begin
	match Vardata.getSubst vd y with
	    None -> x <> y
	  | Some mv -> occurs_check_succeeds vd x mv
      end
  | MData (k,mv) -> occurs_check_succeeds vd x mv
  | MTuple mvs -> List.fold_left
      (fun b mv -> b && occurs_check_succeeds vd x mv) true mvs
  | MAbs (mv,mv') ->
      occurs_check_succeeds vd x mv && occurs_check_succeeds vd x mv'
  | MUnit -> true
  | _ -> failTC "occurs_check"

(* Apply the deterministic constraint-solving rules! *)
let detRules ignoreAbs dec vd cs =

  (* The "progressed" flag keeps a track of whether we've actually
   * done any useful work.
   * If we get to the end and no useful work has been done, we stop.
   * Otherwise, we re-check the leftover constraints in case some
   * of the variables have been substituted away, for example. *)
  let rec loop vd leftover progressed = function

    (* ****************************** *)
    (* NO CONSTRAINTS LEFT TO PROCESS *)
    (* ****************************** *)

    [] -> if progressed then
            (* Re-check the leftover constraints. *)
      loop vd [] false leftover
          else
            (* We've reached the fixpoint, so stop and return! *)
            (vd,leftover)

    (* *************** *)
    (* FRESHNESS RULES *)
    (* *************** *)

    (* Rule F1 *)
    | (FR(x,(ny,ys,MUnit)))::cs -> loop vd leftover true cs
    (* Rule F2 *)
    | (FR(x,(ny,ys,MData(k,mv))))::cs ->
        loop vd leftover true ((FR(x,(ny,ys,mv)))::cs)
    (* Rule F3 *)
    | (FR(x,(ny,ys,MTuple(mvs))))::cs ->
        loop vd leftover true ((List.map (fun mv -> (FR(x,(ny,ys,mv)))) mvs) @ cs)
    (* Freshness for an abstraction - fold into the preceding list! *)
    | (FR(x,(ny,ys,MAbs((MEvar y), mv))))::cs ->
        loop vd leftover true (FR(x,(ny+1,y::ys,mv))::cs)
    (* Freshness for a variable! *)
    | (FR(x,(ny,ys,MEvar y)))::cs ->
        begin
	  if (Types.isNty (Vardata.getType vd y)) then
	    loop vd ((FR(x,(ny,ys,MEvar y)))::leftover) progressed cs
	  else
	    match Vardata.getSubst vd y with
	 	Some mvy ->
		  (* Replace (MEvar y) with mvy, and go back round the loop. *)
		  loop vd leftover true (FR(x,(ny,ys,mvy))::cs)
              | None ->
		  (* Transfer into the "leftover" list.
		   * This DOESN'T affect "progressed", as we'd loop forever! *)
		  loop vd ((FR(x,(ny,ys,MEvar y)))::leftover) progressed cs
	end

    (* ************** *)
    (* EQUALITY RULES *)
    (* ************** *)

    (* Rule E1 *)
    | (EQ((nx,xs,MUnit),(ny,ys,MUnit)))::cs ->
        if nx=ny then loop vd leftover true cs
        else failTC "E1"
    (* Rule E2 *)
    | (EQ((nx,xs,MData(k,mv)),(ny,ys,MData(k',mv'))))::cs ->
        if nx=ny then
          begin
            if k=k' then loop vd leftover true (EQ((nx,xs,mv),(ny,ys,mv'))::cs)
            else raise ConstraintBacktrack
          end
        else failTC "E2"
    (* Rule E3 *)
    | (EQ((nx,xs,MTuple(mvs)),(ny,ys,MTuple(mvs'))))::cs ->
        if nx=ny then
          try
            loop vd leftover true
            ((List.map2 (fun mv mv' -> EQ((nx,xs,mv),(ny,ys,mv'))) mvs mvs')
             @ cs)
          with
            (* In case the tuples have different numbers of elements. *)
            Invalid_argument _ -> failTC "E3a"
        else failTC "E3"
    (* Equality constraints with ONE variable only *)
    | (EQ((nx,xs,MEvar x),(ny,ys,mv)))::cs
    | (EQ((ny,ys,mv),(nx,xs,MEvar x)))::cs when (not(isEvar mv)) ->
        begin
          match Vardata.getSubst vd x with
            Some mvx ->
              (* Replace (MEvar x) with mvx, and go back round the loop. *)
              loop vd leftover true (EQ((nx,xs,mvx),(ny,ys,mv))::cs)
          | None ->
	      (* Need the occurs check here! *)
              if occurs_check_succeeds vd x mv then
		begin
                  if (nx=0 && ny=0) then
                    (* Rule E7 *)
                    let vd' = Vardata.addSubst vd x mv in
                    loop vd' leftover true cs
                  else
                    begin
                      if (nx=ny) then
			(* Transfer to the leftover list.
 			 * This DOESN'T affect "progressed", as we'd loop forever! *)
 			loop vd ((EQ((nx,xs,MEvar x),(ny,ys,mv)))::leftover)
			  progressed cs
		      else
			failTC "equality constraint with only one variable"
		    end
		end
	      else
		raise ConstraintBacktrack
	end
    (* Equality constraints with TWO variables *)
    | (EQ((nx,xs,MEvar x),(ny,ys,MEvar y)))::cs ->
        (* Check whether the abstractions are the same depth! *)
        if nx=ny then
          begin
	    if (Types.isNty (Vardata.getType vd y)) then
	      begin
		if nx=0 && x=y then
		  loop vd leftover true cs
		else
		  loop vd ((EQ((nx,xs,MEvar x),(ny,ys,MEvar y)))::leftover)
		    progressed cs
	      end
	    else
	      match (Vardata.getSubst vd x),(Vardata.getSubst vd y) with
		  Some mvx,Some mvy ->
                    (* Replace (MEvar x) and (MEvar y) with mvx and mvy,
 		     * and go back round the loop. *)
                    loop vd leftover true (EQ((nx,xs,mvx),(ny,ys,mvy))::cs)
		| Some mvx,None ->
                    (* Replace (MEvar x) with mvx, and go back round the loop. *)
                    loop vd leftover true (EQ((nx,xs,mvx),(ny,ys,(MEvar y)))::cs)
		| None,Some mvy ->
                    (* Replace (MEvar y) with mvy, and go back round the loop. *)
                    loop vd leftover true (EQ((nx,xs,(MEvar x)),(ny,ys,mvy))::cs)
		| None,None ->
		    if nx = 0 then
		      begin
			if x=y then
			  (* Simple "x=x" constraint, so delete it. *)
			  loop vd leftover true cs
			else
			  (* This is actually another instance of rule E7. *)
			  let vd' = Vardata.addSubst vd x (MEvar y) in
			  loop vd' leftover true cs
		      end
		    else
		      (* Transfer to the leftover list.
                       * This DOESN'T affect "progressed", as we'd loop forever! *)
		      loop vd ((EQ((nx,xs,MEvar x),(ny,ys,MEvar y)))::leftover)
			progressed cs
	  end
        else
	  failwith "Error in constraint solver - <x1..xm>x = <y1..yn>y where m=/=n"

    (* Two abstractions - fold them into the lists! *)
    (* Rule E4 *)	    
    | (EQ((nx,xs,MAbs((MEvar x),mv)),(ny,ys,MAbs((MEvar y),mv'))))::cs ->
	if ignoreAbs then
	  loop vd leftover true (EQ((0,[],mv),(0,[],mv'))::cs)
	else
          loop vd leftover true (EQ((nx+1,x::xs,mv),(ny+1,y::ys,mv'))::cs)

    (* ALL OTHER POSSIBILITIES SHOULD BE RULED OUT BY THE TYPECHECKER! *)
    | _ -> failTC "detRules"

  in
  loop vd [] false cs

(* Is a given constraint a name constraint? *)
let isNameConstr vd = function
  EQ((_,_,(MEvar x)),_) -> Types.isNty (Vardata.getType vd x)
| FR(_,(_,_,(MEvar y))) -> Types.isNty (Vardata.getType vd y)
| _ -> failwith "isNameConstr - should only have evar-constraints left!"

(* Run a termination check using the deterministic constraint rules themselves! *)
let checkTerm (dec:Types.dtyDec) (vd:Vardata.t) (cs:constr list) =
  (* First of all, filter and process the constraints in the list.
   * Discard any freshness constraints and name-name constraints,
   * and throw away any preceding abstractions. *)
  let cs = linkMap
    (function
	(EQ((nx,xs,(MEvar x)),(ny,ys,(MEvar y)))) ->
	  begin
	    match (Vardata.getType vd x) with
		Nty _ -> []
	      | _ -> [EQ((0,[],(MEvar x)),(0,[],(MEvar y)))]
	  end
      | (EQ((nx,xs,MEvar x),(ny,ys,mv)))
      | (EQ((ny,ys,mv),(nx,xs,MEvar x))) when (not(isEvar mv)) ->
	  [EQ((0,[],(MEvar x)),(0,[],mv))]
      | _ -> []) cs
  in
  (* Now, run the deterministic constraint rules over the
   * modified constraints, IGNORING ABSTRACTIONS.
   * If this fails, we know that there is a cyclic dependency
   * which would cause narrowing to diverge. *)
  begin
    (* Debug output. *)
    debug ("Checking termination::\n" ^
      (string_of_prob string_of_evar vd cs));
    debug ("Current substitution:\n" ^
      (Vardata.string_of_subst string_of_evar soMValue vd));
    (* Actually run the test! *)
    ignore (detRules true dec vd cs)
  end 

(* Function to implement the "narrowing" relation.
 * Needs type information from the datatype declaration and the
 * vardata.
 * This function actually does ALL of the narrowing in one go. *)
let rec narrowAll dec vd ety mv =
  match (mv,ety) with
    (MEvar j, ety) ->
      let (i,vd') = Vardata.addEvar vd ety in
      (MEvar i,vd')
  | (MUnit,UnitTy) -> (MUnit,vd)
  | (MData(k,mv),Dty d) ->
      begin
        match (Types.findDtyCon dec k) with
          Some (ety',d') ->
            if (d=d') then (* Sanity check! *)
	      let (mv',vd') = narrowAll dec vd ety' mv in
	      (MData(k,mv'),vd')
            else failTC "narrow - MData - wrong dty"
        | None -> failTC "narrow - MData - wrong con"
      end
  | (MTuple mvs, TupTy etys) ->
      begin
	try
	  let (mevars,vd') = List.fold_left2
	    (fun (mevars,vd') mv ety ->
	      let (i,vd'') = Vardata.addEvar vd' ety in
	      (mevars@[MEvar i], vd''))
	    ([],vd) mvs etys
	  in
	  (MTuple mevars, vd')
	with Invalid_argument _ ->
	  failTC "narrow - MTuple - values and types don't match"
      end
  | (MAbs((MEvar x),mv), AbsTy(n,ety)) ->
      let (i,vd') = Vardata.addEvar vd (Nty n) in
      let (mv'',vd'') = narrowAll dec vd' ety mv in
      (MAbs((MEvar i),mv''), vd'')
| _ -> failTC "narrow - not possible"

(* Attempt to add some constraints to the current problem,
 * returning the new intermediate constraint representation
 * if the new problem is satisfiable. *)
let addConstraints (dec:Types.dtyDec) (vd:Vardata.t) (cs:prob)
    (ncs:constr list) =
  try

    (* Debug output. *)
    debug ("Testing constraints:\n" ^
      (string_of_prob string_of_evar vd (cs@ncs)));
    debug ("Current substitution:\n" ^
      (Vardata.string_of_subst string_of_evar soMValue vd));

    (* Repeatedly apply the deterministic constraint-solving rules
     * to the new and leftover constraints. *)
    let (vd',cs') = detRules false dec vd (cs@ncs) in

    (* Check for termination. *)
    let _ = checkTerm dec vd' cs' in

    (* Now we know it's OK to narrow, so do all the narrowings
     * in one go. *)
    let (vd'',cs'') = List.fold_left
      (fun (vd',cs') -> function
	| (EQ((nx,xs,MEvar x),(ny,ys,mv)))
	| (EQ((ny,ys,mv),(nx,xs,MEvar x)))
	  when (not(isEvar mv)) && (nx>0) ->
	    let ety = Vardata.getType vd' x in
	    let (mv'',vd'') = narrowAll dec vd' ety mv in
	    let vd'' = Vardata.addSubst vd'' x mv'' in
	    (vd'',cs'@[EQ((nx,xs,mv''),(ny,ys,mv))])
	| c -> (vd',cs'@[c]))
      (vd',[]) cs'
    in

    (* Now we can re-run the determinstic constraint solver.
     * It shouldn't need any more narrowing and it shouldn't
     * diverge (assuming that we did the termination check). *)
    let (vd'',cs'') = detRules false dec vd'' cs'' in

    (* Filter out only those leftover constraints between NAMES. *)
    let namecs = List.filter (isNameConstr vd'') cs'' in

    (* Check satisfiability of the name constraints. *)
    if (Namegraph.checkSat vd'' namecs) then
	(* If successful, return the updated constraint environment. *)
	Some (vd'',cs'')
      else
	(* Otherwise, return None. *)
	None
  with
    (* Constraint solving failed... *)
    ConstraintBacktrack ->
      begin
	debug "Constraint solving failed.";
	None
      end
