(* ************************************* *)
(*                                       *)
(*   alphaML toplevel interactive loop   *)
(*                                       *)
(* ************************************* *)

open Cmd
open Library
open Lexer
open Lexing
open Parser
open Parsing
open Globals
open Location
open Ast
open Errors
open Typesys
open Types
open Machine

(* Display the alphaML logo *)
let display_title () = (let version = "0.3" in print_string
("       _       _           __  __ _     \n" ^
 "  __ _| |_ __ | |__   __ _|  \\/  | |    \n" ^
 " / _` | | '_ \\| '_ \\ / _` | |\\/| | |    \n" ^
 "| (_| | | |_) | | | | (_| | |  | | |___ \n" ^
 " \\__,_|_| .__/|_| |_|\\__,_|_|  |_|_____|\n" ^
 "        |_|                             \n" ^
 "\n" ^
 "Version " ^ version ^ "\n\n"); flush stdout)

(* Display the prompt *)
let display_prompt () = (print_string "> "; flush stdout)

(* An exception for breaking out of the loop *)
exception Quit;;

(* Initialise lexer state, for precise error reporting *)
let lexer_init lexbuf = lexbuf.lex_curr_p <-
  { pos_fname = "toplevel"; pos_lnum = 1; pos_bol = 0; pos_cnum = 0 }

(* Clear parser state after finished a parse *)
let parser_reset () = Parsing.clear_parser ()

(* Parse a single command from lexbuf, taking care of state etc *)
let parse_command lexbuf =
  (lexer_init lexbuf;
   let cmd = Parser.cmd Lexer.token lexbuf in
   parser_reset (); cmd)

(* Parse a list of commands from a file, taking care of state etc *)
let parse_commands filename =
  (let file = open_in filename in
   let lexbuf = Lexing.from_channel file in
   lexer_init lexbuf;
   let cmds = Parser.cmdlist Lexer.token lexbuf in
   parser_reset (); close_in file; cmds)

(* The type of interpreter state! *)
type istate = { mutable st:string list;
	        mutable env:Machine.mvalue list;
	        mutable tenv:(string * ty) list;
	        mutable vd:Vardata.t;
	        mutable cs:Constraint.prob;
	        mutable ndec:ntyDec;
	        mutable dec:dtyDec;
	        mutable rdec:relDec}

(* Use this value to store all the imperative state in one place! *)
let (state : istate) = { st=[]; env=[]; tenv=[]; vd=Vardata.empty;
			 cs=Constraint.trueProb; ndec=[]; dec=[];
			 rdec=[] }

(* Reset the state to its starting values! *)
let reset () =
  begin
    state.st <- []; state.env <- []; state.tenv <- [];
    state.vd <- Vardata.empty; state.cs <- Constraint.trueProb;
    state.ndec <- []; state.dec <- []; state.rdec <- [];
    (* Also reset the state from Globals.ml... *)
    dowalk := true; breadth_first := true; verbose := false
  end

(* Add a value-binding to the relevant places in the state. *)
let add_env x mv t =
  begin
    state.st <- x::state.st;
    state.env <- mv::state.env;
    state.tenv <- (x,t)::state.tenv
  end

(* Add nametype, datatype and relation declarations to the state. *)
let add_ndec new_ndec =
  let ndec' = add_nametypes new_ndec state.dec state.ndec in state.ndec <- ndec'
let add_dec new_dec =
  let dec' = add_datatypes new_dec state.dec state.ndec state.rdec in state.dec <- dec'
let add_rdec new_rdec =
  let rdec' = add_relations new_rdec state.dec state.ndec state.rdec in state.rdec <- rdec'

(* Update the bits of state modified by evaluating an expression. *)
let update_state nvd ncs =
  begin
    state.vd <- nvd;
    state.cs <- ncs
  end

(* Are we processing commands from a file? *)
let filemode = ref false

(* Recursive function to prompt for a "stop/continue" from the user
 * when doing proof-search stuff. *)
let rec stopCarryOn () =
  (output ("Type \',\' (then Enter) to look for more answers, or \'.\' to stop.");
   let lexbuf = Lexing.from_channel stdin in
   try
    (let sco = Parser.searchcmd Lexer.token lexbuf in
     parser_reset (); sco)
   with
     _ -> parser_reset (); stopCarryOn ())

(* Walk over a machine value, to make it more appealing for user output.
 * Evars are interpreted using the function g, which is computed from the
 * results of evaluating the expression. *)
let rec walk vd = function
  MVar i -> failwith "SHOULDN'T TRY TO PRETTY-PRINT MACHINE VARIABLES!"
| MEvar i ->
    begin
      match (Vardata.getSubst vd i) with
	(* Found a substitution! *)
	Some mv' -> walk vd mv'
	(* No substitution. Could try and do something with namegraphs here maybe? *)
      | None -> MEvar i
    end
| MUnit -> MUnit
| MYes -> MYes
| MData (k,mv) -> MData(k,(walk vd mv))
| MTuple mvs -> MTuple(List.map (walk vd) mvs)
| MAbs (mv,mv') -> MAbs((walk vd mv),(walk vd mv'))
| MClosure (f,e) -> MClosure (f,e) (* STAYS THE SAME! *)

(* Look for the variable (if any) that is bound to a particular evar. *)
let lookupEvar env st i =
  match (index (MEvar i) env) with
    None -> None
  | Some i -> Some(List.nth st i)

(* Print an evar! *)
let printEvar env st i =
  match lookupEvar env st i with None -> string_of_evar i | Some x -> x

(* Either walk values before printing them, or don't. *)
let printVal vd mv =
  if !dowalk then soMValueGeneral (printEvar state.env state.st) (walk vd mv)
  else soMValue mv
    (* Something like the following code could be used if we want to
     * return more detailed constraint information with answers.
   let cs = getConstraints () in
   let namecs = List.filter (Solver.isNameConstr vd) cs in
   let solns = Namegraph.enumSolns vd namecs in
   ... *)

(* Process an ExprCmd or LetCmd.
 * This involves parsing, typechecking, expanding and evaluating the expression.
 * The "bindTo" argument says what name to bind the result to, if we care. *)
let rec process_expr re bindTo =

  (* Expand out any defined constructs that might appear! *)
  let re = expand state.dec state.rdec re in

  (* Type check the expanded program. *) 
  let t = typecheck state.dec state.ndec state.rdec state.vd state.tenv re in

  (* Compile it down to abstract machine code. *)
  let amcode = Ast.compile state.dec state.ndec state.st re in

  (* Decide whether to use depth-first or breadth-first mode. *)
  let run = (if !breadth_first then Exec.run_breadth else Exec.run_depth) in

  (* A function to output the appropriate response to the command line. *)
  let respond vd mv =
    match bindTo with
        Some x -> output ("val " ^ x ^ " : " ^ (soType t) ^ " = " ^ (printVal vd mv))
	| None -> output ("- : " ^ (soType t) ^ " = " ^ (printVal vd mv))
  in

  (* A function to update the environment, if necessary. *)
  let maybe_add_env mv =
    match bindTo with
	Some x -> add_env x mv t
      | None -> ()
  in

  (* Recursive function which implements the toplevel interface
   * for dealing with the fact that there may be several answers
   * to a query. *)
  let rec search_eval ss =
    let res = run state.dec ss in
    match res with
      (* Finite failure *)
      None -> output "no"
      (* Found the only answer *)
    | Some (mv,nvd,ncs,[]) ->
        begin
	  update_state nvd ncs;
	  maybe_add_env mv;
          respond nvd mv
        end
      (* Found an answer, but there may be more... *)
    | Some (mv,nvd,ncs,s::ss) ->
        begin
	  respond nvd mv;
          match stopCarryOn () with
              Stop -> begin update_state nvd ncs; maybe_add_env mv end
            | CarryOn -> search_eval (s::ss)
        end
  in
  search_eval [(amcode,[],[state.env],state.vd,state.cs)]

(* The help string which is printed in response to '#help'. *)
let helpstring =
  let lines = ["Interpreter directives:";
	       "%breadth - switch to breadth-first search (the default).";
	       "%constraints - prints current constraint set.";
	       "%debug on/off - toggle printing of debug output (off by default).";
	       "%depth - switch to depth-first search.";
	       "%help - prints this message.";
	       "%quit - exits the interpreter.";
	       "%reset - resets internal state.";
	       "%use \"foo\" - loads file foo and runs all commands therein.";
	       "%walk on/off - toggle walking of values prior to display (on by default)."]
  in
  string_of_list id "\n" lines

(* String representation of booleans, for directives. *)
let sob = function true -> "enabled." | false -> "disabled."

(* Process a command.
 * process_cmd : cmd -> unit *)
let rec process_cmd = function
  Directive d -> process_directive d
| DtyDecCmd new_dec ->
    begin add_dec new_dec; output (rawdec_toString new_dec) end
| NsortDecCmd new_ndec ->
    begin add_ndec new_ndec; output (ndec_toString new_ndec) end
| RelDecCmd new_rdec ->
    begin add_rdec new_rdec; output (rawRelDec_toString new_rdec) end
| ExprCmd(e) -> process_expr e None
| LetCmd(x,xts,e) ->
    let e' = List.fold_right
      (fun (x,t) e_acc -> RawLamExpr(x,t,e_acc,dummy_span))
      xts e
    in
    let xstr = getStr x in
    let bindTo = if (xstr=Ast.wildcard) then None else Some xstr in
    process_expr e' bindTo
| LetRecCmd(f,xts,outty,e) ->
    begin
      match xts with
	[] -> failwith "expand - no arguments for function in let rec"
      | (x,t)::xts' ->
	  let functy = List.fold_right
	    (fun t ty -> RawFunTy(t,ty))
	    (List.map snd xts') outty
	  in
	  let funcbody = List.fold_right
	    (fun (x,t) e_acc -> RawLamExpr(x,t,e_acc,dummy_span))
	    xts' e
	  in
	  let e' = RawRecExpr(f,x,t,functy,funcbody,dummy_span) in
	  process_expr e' (Some(getStr f))
    end

(* Process an interpreter directive.
 * process_directive : directive -> unit *)
and process_directive = function
  VoidDir("quit") -> raise Quit
| VoidDir("help") -> output helpstring
| VoidDir("reset") ->
    begin
      reset ();
      output "Interpreter state has been reset."
    end
| VoidDir("constraints") ->
    begin
      output ("Substitution constraints:");
      output (Vardata.string_of_subst (printEvar state.env state.st)
	(printVal state.vd) state.vd);
      output ("Other constraints:");
      output (Constraint.string_of_prob
		(printEvar state.env state.st)
		state.vd state.cs);
    end
| VoidDir("breadth") ->
    begin
      breadth_first := true;
      output "Breadth-first search enabled."
    end
| VoidDir("depth") ->
    begin
      breadth_first := false;
      output "Depth-first search enabled."
    end
| StringDir("use",filename) ->
    begin
      from_file := true;
      let cmds = parse_commands filename in
      List.iter process_cmd cmds
    end
| BoolDir("walk",b) ->
    begin
      let response =
	"Walking values " ^ (sob b)
      in
      dowalk := b;
      output response
    end
| BoolDir("debug",b) ->
    begin
      let response =
	"Debug output " ^ (sob b)
      in
      verbose := b;
      output response
    end
| VoidDir(x) | StringDir(x,_) | BoolDir(x,_) ->
    raise(UnknownDirective x)

(* Only highlight the errors if not coming from a file! *)
let highlight span = if (!from_file) then () else Location.highlight span

(* Handle an error - handle_error : exn -> unit *)
let handle_error = function
  IllegalToken span ->
    (highlight span;
     output ("Illegal token at " ^ (span_to_string span)))
| SyntaxError span ->
    (highlight span;
     output ("Syntax error at " ^ (span_to_string span)))
| CommentNotTerminated span ->
    (highlight span;
     output ("Comment not terminated at " ^ (span_to_string span)))
| Sys_error msg ->
    output ("Error opening file:" ^ msg)
| Sys.Break -> output "Interrupted."
| WrongType(span,found,expected) ->
    (highlight span;
     output ("Type error at " ^ (span_to_string span));
     output ("Found " ^ (soType found) ^ " but expected " ^ 
       (soTyerr expected)))
| CaseError(span,d) ->
    (highlight span;
     output ("Case error at " ^ (span_to_string span));
     output ("The clauses do not match the datatype " ^ d))
| ProjRange(i,n) ->
    (let span = getSpanInt i in
     highlight span;
     output ("Projection out of range at " ^ (span_to_string span));
     output ("Found " ^ (string_of_int (getInt i)) ^
       ", but should lie in range 1-" ^ (string_of_int n)))
| UnboundVar x ->
    (let span = getSpanId x in
     highlight span;
     output ("Unbound variable " ^ (getStr x) ^ " at " ^
      (span_to_string span)))
| UnboundEvar x ->
    (let span = getSpanId x in
     highlight span;
     output ("Unbound existential variable ?" ^ (getStr x) ^ " at " ^
      (span_to_string span)))
| UnknownCon k ->
    (let span = getSpanId k in
     highlight span;
     output ("Unbound constructor " ^ (getStr k) ^ " at " ^
      (span_to_string span)))
| AlreadyDty x ->
    output ("Type identifier " ^ x ^ " already defined as a datatype.")
| AlreadyNty x ->
    output ("Type identifier " ^ x ^ " already defined as a nametype.")
| AlreadyCon x ->
    output ("Data constructor " ^ x ^ " already defined.")
| AlreadyRel x ->
    output ("Relation constructor " ^ x ^ " already defined.")
| UnboundTypeId x ->
    output ("Unbound type identifier " ^ x)
| AbsError x ->
    output ("Cannot abstract over " ^ x ^ " as not a nametype.")
| NotEqTy (sp,t) ->
    (highlight sp;
     output ("Expected an equality type at " ^
       (span_to_string sp)))
| NotInhab (sp,t) ->
    (highlight sp;
     output ("Expected a non-empty type at " ^
       (span_to_string sp)))
| NotNty(n) ->
    (let sp = getSpanId n in
     highlight sp;
     output ("Expected a nametype at " ^ (span_to_string sp)))
| UnknownDirective x ->
    output ("Unknown directive %" ^ x)
| e -> raise e

(* The actual toplevel loop *)
let loop () =
  display_title ();
  Sys.catch_break true; (* Control-C interrupts current evaluation *)
  try
    while true do
      (from_file := false;
       display_prompt ();
       try
         let lexbuf = Lexing.from_channel stdin in
         let parsed_input = parse_command lexbuf in
         process_cmd parsed_input
       with
         e -> handle_error e)
     done
  with
    Quit -> output "Goodbye!"
