open Library

(* ********************************************** *)
(*                                                *)
(*   REPRESENTATION OF ALPHAML TYPES              *)
(*     The interpreter carries some type info,    *)
(*     basically so that the toplevel can be      *)
(*     informed about the types of dynamically-   *)
(*     generated variables.                       *)   
(*                                                *)
(* ********************************************** *)

(* alphaML types - with type identifiers fully interpreted *)
type ty =
  UnitTy
| PropTy
| Nty   of string
| Dty   of string
| TupTy of ty list
| FunTy of ty * ty
| AbsTy of string * ty
| RelTy

(* Is something a name type? *)
let isNty = function Nty _ -> true | _ -> false

(* to_string function for types *)
let rec soType x =
 let brackets x = match x with
   TupTy _ | FunTy _ -> "(" ^ (soType x) ^ ")"
 | _ -> soType x
 in match x with
  UnitTy -> "unit"
| PropTy -> "prop"
| Nty n -> n
| TupTy ts -> string_of_list brackets " * " ts
| Dty d -> d
| FunTy (t,t') -> (brackets t) ^ " -> " ^ (brackets t')
| AbsTy (x,rt) -> "[" ^ x ^ "]" ^ (brackets rt)
| RelTy -> "rel"

(* alphaML nametype and datatype declarations
 * (fully interpreted) *)
type ntyDec = string list
type dtyDec = (string * ((string * ty) list)) list
type relDec = (string * ty) list

(* to_string function for nametype declarations *)
let ndec_toString (ndec:ntyDec) =
  ("nametype " ^ (string_of_list id "\n     and " ndec))

(* Is this type an equality type, under the given datatype declaration?
 * Assumes that all abstraction types are actually over nametypes. *)
let rec isEqTy dec lst t =
  let areEqTys lst ts = List.for_all (isEqTy dec lst) ts in
  match t with
    UnitTy -> true
  | PropTy -> false
  | Nty n -> true
  | Dty d -> (List.mem d lst) ||
      (match get2 d dec with
        None -> failwith "Error calculating equality type"
      | Some kts -> areEqTys (d::lst) (List.map snd kts))
  | TupTy ts -> areEqTys lst ts
  | FunTy (t,t') -> false
  | AbsTy (n,t) -> isEqTy dec lst t
  | RelTy -> true

(* Given a datatype declaration and a list of nominal data sorts
 * (Dty d) which are inhabited, work out whether the given type
 * is inhabited. *)
let rec isInhab dec lst badlst t =
  let allInhab lst badlst ts = List.for_all (isInhab dec lst badlst) ts in
  let someInhab lst badlst ts = List.exists (isInhab dec lst badlst) ts in
  match t with
    UnitTy -> true
  | PropTy -> false
  | Nty n -> true
  | Dty d -> (not(List.mem d badlst)) && ((List.mem d lst) ||
      (match get2 d dec with
        None -> failwith "Error calculating inhabited types"
      | Some kts -> someInhab lst (d::badlst) (List.map snd kts)))
  | TupTy ts -> allInhab lst badlst ts
  | FunTy (t,t') -> false
  | AbsTy (n,t) -> isInhab dec lst badlst t
  | RelTy -> false

(* Return the list of all inhabited nominal data sorts in a
 * given datatype declaration.
 * This means that all possible argument types must
 * be equality types and must themselved be inhabited.
 * This function uses a fixpoint computation,
 * which must terminate as it is over a finite domain. *)
let inhabDtys (dec:dtyDec) =
  let all_dtys = List.map fst dec in
  let rec find_fixpt lst =
    let next_step lst =
      List.filter (fun d -> (List.mem d lst) ||
        (match get2 d dec with
           None -> failwith "Error calculating inhabited datatypes"
         | Some kts ->
             List.for_all (isInhab dec lst []) (List.map snd kts))) all_dtys
    in
    let lst' = next_step lst in
    if (lst=lst') then lst' else find_fixpt lst'
  in
  find_fixpt []

(* Find a particular constructor in a datatype declaration.
 * Returns the argument and target types *)
let findDtyCon (dec:dtyDec) (k:string) =
  let rec inner = function
    [] -> None
  | (d,cs)::dec -> (match get2 k cs with
      None -> inner dec
    | Some ty -> Some (ty,d))
  in inner dec

(* Find a particular constructor in a relation declaration,
 * Returns the argument and target types *)
let findRelCon (rdec:relDec) (r:string) =
  let rec inner = function
    [] -> None
  | (r',t)::rdec -> if (r=r') then Some t else inner rdec
  in
  inner rdec
