open Ast
open Library
open Location
open Types
open Errors

(* ********************************************************* *)
(*                                                           *)
(*       TYPECHECKER OPERATING OVER RAW EXPRESSIONS          *)
(*        (SO WE CAN RETURN HELPFUL ERROR MESSAGES)          *)
(*                                                           *)
(* ********************************************************* *)

(* Add a nametype declaration to the current environment *)
let add_nametypes (new_ndec:ntyDec) (dec:dtyDec) (ndec:ntyDec) =
  let existing_dtys = List.map fst dec in
  match (findDup (new_ndec@ndec)),
        (findDup (new_ndec@existing_dtys))
  with
    None,None -> (ndec@new_ndec)
  | (Some n),_ -> raise (AlreadyNty n)
  | _,(Some d) -> raise (AlreadyDty d)

(* Add a datatype declaration to the current environment *)
let add_datatypes (new_dec:rawDtyDec) (dec:dtyDec) (ndec:ntyDec) (rdec:relDec) =
  let existing_dtys = List.map fst dec in
  let new_dtys = List.map fst new_dec in
  let new_rhsides = linkMap snd new_dec in
  let existing_rhsides = linkMap snd dec in
  let new_cons = List.map fst new_rhsides in
  let existing_cons = List.map fst existing_rhsides in
  let existing_rels = List.map fst rdec in
  (* Check for possible errors *)
  let dup_new_dty = findDup new_dtys in
  let dup_existing_dty = findDup (new_dtys@existing_dtys) in
  let dup_existing_nty = findDup (new_dtys@ndec) in
  let dup_new_con = findDup new_cons in
  let dup_existing_con = findDup (new_cons@existing_cons) in
  let dup_existing_rel = findDup (new_cons@existing_rels) in
  match dup_new_dty,dup_existing_dty,dup_existing_nty,
        dup_new_con,dup_existing_con,dup_existing_rel with
    Some d,_,_,_,_,_ -> raise (AlreadyDty d)
  | _,Some d,_,_,_,_ -> raise (AlreadyDty d)
  | _,_,Some n,_,_,_ -> raise (AlreadyNty n)
  | _,_,_,Some k,_,_ -> raise (AlreadyCon k)
  | _,_,_,_,Some k,_ -> raise (AlreadyCon k)
  | _,_,_,_,_,Some r -> raise (AlreadyRel r)
  | None,None,None,None,None,None ->
      let classify = classifyIds ndec (existing_dtys@new_dtys) in
      let new_dec' = List.map
        (fun (d,kts) -> (d,(List.map
          (fun (k,t) ->
             let trnt = trnRawType classify t in
             (k,trnt))
           kts)))
         new_dec
      in
      (dec@new_dec')

(* Add a relation declaration to the current environment *)
let add_relations (new_rdec:rawRelDec) (dec:dtyDec) (ndec:ntyDec) (rdec:relDec) =
  let existing_dtys = List.map fst dec in
  let new_rels = List.map fst new_rdec in
  let existing_rels = List.map fst rdec in
  let existing_rhsides = linkMap snd dec in
  let existing_cons = List.map fst existing_rhsides in
  (* Check for possible errors *)
  let dup_new_rel = findDup new_rels in
  let dup_existing_rel = findDup (new_rels@existing_rels) in
  let dup_existing_con = findDup (new_rels@existing_cons) in
  match dup_new_rel,dup_existing_rel,dup_existing_con with
    Some r,_,_ -> raise (AlreadyRel r)
  | _,Some r,_ -> raise (AlreadyRel r)
  | _,_,Some k -> raise (AlreadyCon k)
  | None,None,None ->
      let classify = classifyIds ndec (existing_dtys) in
      let new_rdec' =
        List.map (fun (r,rt) -> (r,trnRawType classify rt)) new_rdec
      in
      (rdec@new_rdec')

(* MAIN TYPECHECKING FUNCTION FOR EXPRESSIONS. *)
let typecheck (dec:dtyDec) (ndec:ntyDec) (rdec:relDec)
  (evarenv:Vardata.t) (tenv:(string*ty) list) (e:rawExpr) =

  (* A function for returning typechecking errors. *)
  let tc_error str = failwith ("Typechecking error - " ^ str) in
  (* A function for classifying type indentifiers *)
  let classify = classifyIds ndec (List.map fst dec) in
  (* Computes the list of inhabited nominal data sorts (Dty d) in dec. *)
  let inhabd = inhabDtys dec in

  (* Inner function for typechecking an expression *)
  let rec tcExpr tenv = function
    RawVarExpr x ->
      begin
	match (get2 (getStr x) tenv) with
           Some ty -> ty
         | None -> raise (UnboundVar x)
      end
  | RawEvarExpr x ->
      begin
        try
          Vardata.getType evarenv (int_of_xyzstring (getStr x))
        with
          _ -> raise (UnboundEvar x)
      end
  | RawLamExpr (x,ty,e,sp) ->
      let trnty = trnRawType classify ty in
      let ty' = tcExpr (((getStr x),trnty)::tenv) e in
      FunTy(trnty,ty')
  | RawRecExpr (f,x,inty,outty,e,sp) ->
      let trninty = trnRawType classify inty in
      let trnoutty = trnRawType classify outty in
      let tenv' = ((getStr x),trninty)::
        ((getStr f),(FunTy(trninty,trnoutty)))::tenv
      in
      let t' = tcExpr tenv' e in
      if (t' = trnoutty) then
        FunTy(trninty,trnoutty)
      else raise (WrongType((getSpanExpr e),t',TYPE trnoutty))
  | RawAppExpr (es,sp) ->
     (* Helper function to get the types right - need to
      * reverse input list first though! *)
      let rec docheck = function
        [] -> tc_error " empty application node"
      | [e] -> tc_error "singleton application node"
      | [e';e] ->
          (match tcExpr tenv e with
             FunTy(t,t') ->
               let t'' = tcExpr tenv e' in
               if (t'' = t) then t' else
               raise (WrongType((getSpanExpr e'),t'',TYPE t))
           | ty -> raise
               (WrongType((getSpanExpr e),ty,DESCR "a function type")))
      | e::es ->
          (let esspan = 
             let start = getSpanExpr (final es) in
             let finish = getSpanExpr (List.hd es) in
             mergeSpans start finish
           in
           match docheck es with
             FunTy(t,t') ->
               let ty = tcExpr tenv e in
               if (ty = t) then t' else
               raise (WrongType((getSpanExpr e),ty,TYPE t))
           | ty -> raise (WrongType(esspan,ty,DESCR "a function type")))
      in
      docheck (List.rev es)
  | RawProjExpr (e,i,sp) ->
      (let iv = getInt i in
       match tcExpr tenv e with
         TupTy ts ->
          let n = List.length ts in
          if (1 <= iv && iv <= n) then
            List.nth ts (iv-1)
          else
            raise (ProjRange (i,n))
      | ty -> raise (WrongType((getSpanExpr e),ty,DESCR "a tuple type")))
  | RawCaseExpr (e,cs,sp) ->
      (match tcExpr tenv e with
       Dty d ->
         (match get2 d dec with
          Some kts ->
            let case_error = CaseError(sp,d) in
            if not(List.length cs = List.length kts) then
              raise case_error
            else
              (let sorted_cs = List.sort
                 (fun x y -> compare (fst3 x) (fst3 y)) cs
               in
               let sorted_kts = List.sort
                 (fun x y -> compare (fst x) (fst y)) kts
               in
               if not(List.map (fun x -> (getStr(fst3 x))) sorted_cs =
                      List.map fst sorted_kts)
               then
                 raise case_error
               else
                 let arities = List.map snd sorted_kts in
                 (match sorted_cs,arities with
                   (k1,x1,e1)::cs, t1::arities ->
                     let t =
                       tcExpr (((getStr x1),t1)::tenv) e1
                     in
                     List.fold_left2
                       (fun t (ki,xi,ei) ti ->
                        let t' =
                          tcExpr (((getStr xi),ti)::tenv) ei
                        in
                        if t=t' then t else
                        raise (WrongType((getSpanExpr ei),t',TYPE t)))
                    t cs arities
  	       | _ -> tc_error "\'case\' statement (1)"))
        | _ -> tc_error "\'case\' statement (2)")
     | ty -> raise (WrongType((getSpanExpr e),ty,DESCR "a datatype")))
  | RawDataExpr (k,e,sp) ->
      begin
	match findDtyCon dec (getStr k) with
          Some (t,d) -> let t' = tcExpr tenv e in
            if t'=t then (Dty d) else
            raise (WrongType((getSpanExpr e),t',TYPE t))
        | None ->
	    begin
	      match findRelCon rdec (getStr k) with
		Some t -> let t' = tcExpr tenv e in
		  if t'=t then RelTy else
		  raise (WrongType((getSpanExpr e),t',TYPE RelTy))
	      | None -> raise (UnknownCon k)
	    end
      end
  | RawTupExpr (es,sp) -> TupTy (List.map (tcExpr tenv) es)
  | RawUnitExpr sp -> UnitTy
  | RawSomeExpr (t,sp) ->
      let trnt = trnRawType classify t in
      (match (isEqTy dec [] trnt),(isInhab dec inhabd [] trnt) with
         false,_ -> raise (NotEqTy (sp,trnt))
       | _,false -> raise (NotInhab (sp,trnt))
       | true,true -> trnt)
  | RawFreshExpr (n,sp) ->
      let nstr = getStr n in
      if (List.mem nstr ndec) then (Nty nstr)
      else raise(NotNty n)
  | RawAbsExpr (e,e',sp) ->
      (match tcExpr tenv e with
         Nty n ->
           let t = tcExpr tenv e' in
           if (isEqTy dec [] t) then AbsTy(n,t)
           else raise (NotEqTy ((getSpanExpr e'),t))
       | ty -> raise (WrongType((getSpanExpr e),ty,DESCR "a nametype")))
  | RawBranchExpr (es,sp) ->
      (match es with
         [] -> tc_error "no branches."
       | e::es ->
          (let t = tcExpr tenv e in
           List.iter (fun e' -> let t' = tcExpr tenv e' in
             if (t=t') then () else
             raise(WrongType ((getSpanExpr e'),t',TYPE t))) es; t))
  | RawSeqExpr (e,e',sp) ->
      let _ = tcExpr tenv e in tcExpr tenv e'
  | RawFailExpr (t,sp) -> trnRawType classify t
  | RawSuccessExpr sp -> PropTy
  | RawEconExpr (e,e',sp) ->
      let t = tcExpr tenv e in
      let isEqTyT = isEqTy dec [] t in
      let t' = tcExpr tenv e' in
      let isEqTyT' = isEqTy dec [] t' in
      (match isEqTyT,isEqTyT',t=t' with
        false,_,_ -> raise(NotEqTy ((getSpanExpr e),t))
      | _,false,_ -> raise(NotEqTy ((getSpanExpr e'),t'))
      | _,_,false -> raise(WrongType ((getSpanExpr e'),t',TYPE t))
      | true,true,true -> PropTy)
  | RawFconExpr (e,e',sp) ->
      (match tcExpr tenv e with
         Nty n ->
           let t = tcExpr tenv e' in
           if (isEqTy dec [] t) then PropTy
           else raise (NotEqTy ((getSpanExpr e'),t))
       | ty -> raise (WrongType((getSpanExpr e),ty,DESCR "a nametype")))
  | RawDistinctExpr (es,sp) ->
      let _ = List.iter
	(fun e -> let t = tcExpr tenv e in
	 if isNty t then ()
         else raise (WrongType((getSpanExpr e),t,DESCR "a nametype")))
	es
      in PropTy
  | RawLetExpr (x,xts,e,e',sp) ->
      begin
	match xts with
	    [] -> let ty = tcExpr tenv e in
                  tcExpr (((getStr x),ty)::tenv) e'
	  | _ -> tc_error "non-trivial let expressionss should have been expanded"
      end
  | RawPrintExpr _ -> UnitTy (* ?????????? *)	
  | RawDefnExpr _ -> tc_error "inductive definitions should have been expanded"
  | RawLetRecExpr _ -> tc_error "let rec expressions should have been expanded"
  | RawUnbindExpr _ -> tc_error "unbind expressions should have been expanded"
  | RawExistsExpr _ -> tc_error "exists expressions should have been expanded"
  | RawTreeExpr _ -> tc_error "tree expressions should have been expanded"
  in
  tcExpr tenv e
