open Library
open Types
open Machine

(* Set a particular element of an ('a option) Vec.t.
 * If the index is out of range, the vector is repeatedly
 * DOUBLED in size until it is in range. *)
let rec set_autogrow i d v =
  let len = Vec.length v in
  let grow_double v =
    let v' = ref (Vec.create None (len * 2)) in
    let _ = Vec.iteri (fun i d -> v' := Vec.set i d !v') v in
    !v'
  in
  if (len > i) then
    Vec.set i d v
  else
    set_autogrow i d (grow_double v)

(* Store information about variables - their type and (maybe) a substitution. *)
type t = { nextEvar: int;
           data: ((ty * (mvalue option)) option) Vec.t}

let empty = { nextEvar = 0; data = Vec.singleton None }

let addEvar vd ety =
  let i = vd.nextEvar in
  let data' = set_autogrow i (Some (ety,None)) vd.data in
  (i, { nextEvar = i+1; data = data' })

let addEvarWithSubst vd ety mv =
  let i = vd.nextEvar in
  let data' = set_autogrow i (Some (ety,Some mv)) vd.data in
  (i, { nextEvar = i+1; data = data' })

let addSubst vd i mv =
  try
    match (Vec.get i vd.data) with
      Some(ty,None) -> 
        let data' = Vec.set i (Some (ty,Some mv)) vd.data in
        { nextEvar = vd.nextEvar; data = data' }
    | Some(ty,Some _) -> failwith "addSubst - evar already has a substitution"
    | None -> failwith "addSubst - evar doesn't exist"
  with
    Vec.Vec_index_out_of_bounds -> failwith "addSubst - evar doesn't exist"

let getSubst vd i =
  try
    match (Vec.get i vd.data) with
      Some(_,s) -> s
    | None -> failwith "getSubst - evar doesn't exist"
  with
    Vec.Vec_index_out_of_bounds -> failwith "getSubst - evar doesn't exist"

let getType vd i =
  try
    match (Vec.get i vd.data) with
      Some(ety,_) -> ety
    | None -> failwith "getType - evar doesn't exist"
  with
    Vec.Vec_index_out_of_bounds -> failwith "getType - evar doesn't exist"

let lastEvar vd = vd.nextEvar - 1

let string_of_subst printevar printvalue vd = 
  let rev_print = Vec.foldi
    (fun i x acc -> match x with
	Some (ty,(Some mv)) -> (i,mv)::acc
      | _ -> acc)
    vd.data []
  in
  string_of_enclist "{" "}"
    (fun (i,mv) -> ((printevar i) ^ " = " ^ (printvalue mv)))
    ",\n " (List.rev rev_print)

let get_mappings vd =
  Vec.foldi (fun i x acc -> 
    match x with None -> acc
      | Some(t,None) -> acc
      | Some(t,(Some mv)) -> (i,mv)::acc) vd.data []
